# Copyright 2023-2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Craft-application based entrypoint for charmcraft.

This submodule contains what's needed to run charmcraft with the
craft-application framework, falling back to "classic" charmcraft when a command
is not available.

Help is (for now) handled by classic charmcraft.
"""

from charmcraft.application.main import APP_METADATA, Charmcraft, main, get_app_info

__all__ = ["APP_METADATA", "Charmcraft", "main", "get_app_info"]



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Service class for interacting with charm libraries."""

from __future__ import annotations

import dataclasses
import pathlib
from collections.abc import Collection
from typing import cast

import craft_application

from charmcraft import utils
from charmcraft.services.store import StoreService
from charmcraft.store.models import Library


@dataclasses.dataclass
class CharmLibDelta:
    """The difference between a local charmlib and a store charmlib."""

    lib_name: str
    local_version: tuple[int, int] | None
    store_version: tuple[int, int] | None


class CharmLibsService(craft_application.AppService):
    """Business logic for dealing with charm libraries."""

    def __init__(
        self,
        app: craft_application.AppMetadata,
        services: craft_application.ServiceFactory,
        *,
        project_dir: pathlib.Path,
    ) -> None:
        super().__init__(app, services)
        self._project_dir = project_dir

    def is_downloaded(
        self, *, charm_name: str, lib_name: str, api: int, patch: int | None = None
    ) -> bool:
        """Check if the given charm lib is already downloaded on disk.

        :param charm_name: The name of the charm the lib is attached to.
        :param lib_name: The name of the lib itself.
        :param api: The api version of the lib
        :param patch: If given, the specific patch version of the lib.
        """
        lib_path = utils.get_lib_path(charm_name, lib_name, api)
        if not (self._project_dir / lib_path).exists():
            return False

        if patch is None:
            return True

        lib_info = utils.get_lib_info(lib_path=self._project_dir / lib_path)
        return lib_info.patch == patch

    def get_local_version(
        self, *, charm_name: str, lib_name: str
    ) -> tuple[int, int] | None:
        """Get the version of the library on the machine, or None.

        :param charm_name: The name of the charm where the lib is published
        :param lib_name: The name of the library itself
        :returns: Either the version of the library as a pair of integers or None
            if the library cannot be found.
        """
        charm_libs_path = self._project_dir / utils.get_lib_charm_path(charm_name)
        if not charm_libs_path.is_dir():
            return None
        for api_version_path in charm_libs_path.iterdir():
            lib_path = api_version_path / f"{lib_name}.py"
            if lib_path.exists() and lib_path.is_file() or lib_path.is_symlink():
                info = utils.get_lib_info(lib_path=lib_path)
                if info.patch == -1:
                    return None
                return (info.api, info.patch)
        return None

    def write(self, library: Library) -> None:
        """Write the given library to disk.

        :param library: A store library object with valid content.
        """
        if library.content is None:
            # This should be considered an internal error.
            raise ValueError("Library has no content.")
        lib_path = self._project_dir / utils.get_lib_path(
            library.charm_name, library.lib_name, library.api
        )
        lib_path.parent.mkdir(parents=True, exist_ok=True)
        lib_path.write_text(library.content)

    def get_unpublished_libs(self) -> Collection[CharmLibDelta]:
        """Get this charm's unpublished charmlibs.

        Get the charmlibs owned by this charm that are newer on disk than on Charmhub.
        """
        project_name = self._services.get("project").get().name
        local_libs = utils.get_libs_from_tree(project_name, self._project_dir)

        store = cast(StoreService, self._services.get("store"))
        store_libs = store.get_libraries_metadata_by_id(
            *(lib.lib_id for lib in local_libs if lib.lib_id)
        )
        unpublished_libs = []
        for lib in local_libs:
            local_version = (lib.api, lib.patch)
            store_lib = store_libs.get(str(lib.lib_id))
            store_version = (store_lib.api, store_lib.patch) if store_lib else None
            if store_version is None or local_version > store_version:
                unpublished_libs.append(
                    CharmLibDelta(
                        lib_name=lib.lib_name,
                        local_version=local_version,
                        store_version=store_version,
                    )
                )

        return unpublished_libs



# Copyright 2021-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Craft-parts setup, lifecycle and plugins."""

from typing import Any

import craft_parts
from craft_parts.parts import PartSpec

from . import plugins
from charmcraft.parts.lifecycle import PartsLifecycle

__all__ = [
    "plugins",
    "get_app_plugins",
    "setup_parts",
    "process_part_config",
    "PartsLifecycle",
]


def get_app_plugins() -> dict[str, type[craft_parts.plugins.Plugin]]:
    """Get the app-specific plugins for Charmcraft."""
    return {
        "charm": plugins.CharmPlugin,
        "poetry": plugins.PoetryPlugin,
        "python": plugins.PythonPlugin,
        "reactive": plugins.ReactivePlugin,
        "uv": plugins.UvPlugin,
    }


def setup_parts() -> None:
    """Initialize craft-parts plugins."""
    craft_parts.plugins.register(get_app_plugins())


def process_part_config(data: dict[str, Any]) -> dict[str, Any]:
    """Validate and fill the given part data against/with common and plugin models.

    :param data: The part data to use.

    :return: The part data validated and completed with plugin defaults.
    """
    if not isinstance(data, dict):
        raise TypeError("value must be a dictionary")

    # copy the original data, we'll modify it
    spec = data.copy()

    plugin_name = spec.get("plugin")
    if not plugin_name:
        raise ValueError("'plugin' not defined")

    plugin_class = craft_parts.plugins.get_plugin_class(plugin_name)

    # validate plugin properties
    plugin_properties = plugin_class.properties_class.unmarshal(spec)

    # validate common part properties
    part_spec = craft_parts.plugins.extract_part_properties(
        spec, plugin_name=plugin_name
    )
    PartSpec(**part_spec)

    # get plugin properties data if it's model based (otherwise it's empty), and
    # update with the received config
    if isinstance(plugin_properties, craft_parts.plugins.PluginProperties):
        full_config = plugin_properties.model_dump(by_alias=True, exclude_unset=True)
    else:
        full_config = {}
    full_config.update(data)

    return full_config



#!/usr/bin/env python3
# Copyright 2025 root
# See LICENSE file for licensing details.

"""Django Charm entrypoint."""

import logging
import typing

import ops

import paas_charm.django

logger = logging.getLogger(__name__)


class DjangoHelloWorldCharm(paas_charm.django.Charm):
    """Django Charm service."""

    def __init__(self, *args: typing.Any) -> None:
        """Initialize the instance.

        Args:
            args: passthrough to CharmBase.
        """
        super().__init__(*args)
        self.framework.observe(self.on.clearsession_action,
                               self._on_clearsession_action)

    def _on_clearsession_action(self, event: ops.ActionEvent) -> None:
        """Handle the clearsession action.
        Args:
            event: the action event object.
        """
        if not self.is_ready():
            event.fail("django-app container is not ready")
        try:
            self._container.exec(
                ["python3", "manage.py", "clearsessions"],
                service_context="django",
                combine_stderr=True,
                working_dir=str(self._workload_config.app_dir),
            ).wait_output()
            event.set_results({"result": "session cleared!"})
        except ops.pebble.ExecError as e:
            event.fail(str(e.stdout))

if __name__ == "__main__":
    ops.main(DjangoHelloWorldCharm)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Constants used in charmcraft."""

import enum
from typing import Literal

from craft_platforms import DistroBase
from craft_providers.bases import BaseName

# region Environment variables
ALTERNATE_AUTH_ENV_VAR = "CHARMCRAFT_AUTH"
DEVELOPER_MODE_ENV_VAR = "CHARMCRAFT_DEVELOPER"
EXPERIMENTAL_EXTENSIONS_ENV_VAR = "CHARMCRAFT_ENABLE_EXPERIMENTAL_EXTENSIONS"
IMAGE_INFO_ENV_VAR = "CHARMCRAFT_IMAGE_INFO"
PROVIDER_ENV_VAR = "CHARMCRAFT_PROVIDER"
SHARED_CACHE_ENV_VAR = "CRAFT_SHARED_CACHE"
STORE_API_ENV_VAR = "CHARMCRAFT_STORE_API_URL"
STORE_STORAGE_ENV_VAR = "CHARMCRAFT_UPLOAD_URL"
STORE_REGISTRY_ENV_VAR = "CHARMCRAFT_REGISTRY_URL"
# These are only for use within the managed environment
MANAGED_MODE_ENV_VAR = "CHARMCRAFT_MANAGED_MODE"
# endregion
# region Project files and directories
CHARMCRAFT_FILENAME = "charmcraft.yaml"
MANIFEST_FILENAME = "manifest.yaml"
JUJU_CONFIG_FILENAME = "config.yaml"
METADATA_FILENAME = "metadata.yaml"
JUJU_ACTIONS_FILENAME = "actions.yaml"

WORK_DIRNAME = "work_dir"
BUILD_DIRNAME = "build"
VENV_DIRNAME = "venv"
STAGING_VENV_DIRNAME = "staging-venv"
# endregion
# region Output files and directories
# Dispatch script filename
DISPATCH_FILENAME = "dispatch"
# Hooks directory name
HOOKS_DIRNAME = "hooks"
# The minimum set of hooks to be provided for compatibility with old Juju
MANDATORY_HOOK_NAMES = frozenset(("install", "start", "upgrade-charm"))

LEGACY_BASES = (  # Legacy bases that can use the "bases" syntax.
    "ubuntu@18.04",
    "ubuntu@20.04",
    "ubuntu@22.04",
    "almalinux@9",
)

CHARM_OR_REACTIVE_BASES = frozenset(  # Bases with the 'charm' & 'reactive' plugins.
    (
        *LEGACY_BASES,
        "ubuntu@24.04",
        "ubuntu@24.10",
        "ubuntu@25.04",
    )
)

CommonBaseStr = Literal[  # Bases supported as both build bases and run bases
    "ubuntu@18.04",
    "ubuntu@20.04",
    "ubuntu@22.04",
    "ubuntu@24.04",
    "ubuntu@25.04",
    "ubuntu@25.10",
    "almalinux@9",
]
BaseStr = CommonBaseStr | Literal["ubuntu@26.04"]
BuildBaseStr = CommonBaseStr | Literal["ubuntu@devel"]

DEVEL_BASE_STRINGS = (
    "ubuntu@25.04",
    "ubuntu@25.10",
    "ubuntu@26.04",
)  # Bases that require a specified build base.

SUPPORTED_BASE_STRINGS = frozenset(
    # Bases that do not need a specific "devel" build base.
    (
        *(
            f"ubuntu@{series}"
            for series in ("18.04", "20.04", "22.04", "24.04", "25.04", "25.10")
        ),
        *(("almalinux@9",)),
    )
)

SUPPORTED_BASES = frozenset(
    BaseName(base.distribution, base.series)
    for base in (DistroBase.from_str(base_str) for base_str in SUPPORTED_BASE_STRINGS)
)

SUPPORTED_OSES = frozenset(base.name for base in SUPPORTED_BASES)


class CharmArch(str, enum.Enum):
    """An architecture for a charm."""

    amd64 = "amd64"
    arm64 = "arm64"
    armhf = "armhf"
    ppc64el = "ppc64el"
    riscv64 = "riscv64"
    s390x = "s390x"

    def __str__(self) -> str:
        return str(self.value)


GO_ARCH_TO_CHARM_ARCH = {
    "arm": "armhf",
    "ppc64le": "ppc64el",
}
"""Mapping to convert go architectures to charm architecture strings.

Architectures not included here are the same in GOARCH as charm arch names.

go architectures are also used as OCI image architectures.
Reference 1: https://github.com/opencontainers/image-spec/blob/main/config.md#properties
Reference 2: https://go.dev/doc/install/source#environment
"""


SUPPORTED_ARCHITECTURES = frozenset(arch.value for arch in CharmArch)


# The minimum set of files for a charm to be considered valid
CHARM_MANDATORY_FILES = frozenset(
    (
        DISPATCH_FILENAME,
        HOOKS_DIRNAME,
    )
)
# Optional files that can be present in a charm
CHARM_OPTIONAL_FILES = frozenset(
    (
        METADATA_FILENAME,
        JUJU_ACTIONS_FILENAME,
        JUJU_CONFIG_FILENAME,
        "metrics.yaml",
        "lxd-profile.yaml",
        "templates",
        "version",
        "lib",
        "mod",
        "LICENSE",
        "icon.svg",
        "README.md",
        "actions",
    )
)
# endregion

DEPENDENCIES_HASH_FILENAME = "charmcraft-dependencies-hash.txt"

# If Juju doesn't support the dispatch mechanism, it will execute the
# hook, and we'd need sys.argv[0] to be the name of the hook but it's
# getting lost by calling this dispatch, so we fake JUJU_DISPATCH_PATH
# to be the value it would've otherwise been.
DISPATCH_CONTENT = """#!/bin/sh

JUJU_DISPATCH_PATH="${{JUJU_DISPATCH_PATH:-$0}}" PYTHONPATH=lib:venv \\
  exec ./{entrypoint_relative_path}
"""

UBUNTU_LTS_STABLE = frozenset(
    (
        "18.04",
        "20.04",
        "22.04",
        "24.04",
    )
)

# Metadata keys that are defined in the metadata.yaml file, for backwards compatible
METADATA_YAML_KEYS = frozenset(
    (
        "assumes",
        "containers",
        "description",
        "devices",
        "display-name",
        "docs",
        "extra-bindings",
        "issues",
        "maintainers",
        "name",
        "peers",
        "provides",
        "requires",
        "resources",
        "series",
        "storage",
        "subordinate",
        "summary",
        "terms",
        "website",
    )
)

CHARM_METADATA_LEGARY_KEYS_ALIAS = frozenset(
    (
        "display_name",
        "extra_bindings",
    )
)

# Metadata keys that are allowed in the charmcraft.yaml file
CHARM_METADATA_KEYS = frozenset(
    (
        "assumes",
        "containers",
        "description",
        "devices",
        "title",
        "documentation",
        "extra-bindings",
        "links",
        "name",
        "peers",
        "provides",
        "requires",
        "resources",
        "storage",
        "subordinate",
        "summary",
        "terms",
    )
)

CHARM_METADATA_KEYS_ALIAS = frozenset(("extra_bindings",))

METADATA_YAML_MIGRATE_FIELDS = ("name", "summary", "description")
"""Fields that can exist in metadata.yaml or charmcraft.yaml, but not both."""



# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.

"""Smoke scenario test for paas-charm based init templates."""

import os
import pathlib

import charm
import scenario
import scenario.errors


def test_smoke():
    """The purpose of this test is that the charm does not raise on a handled event."""
    os.chdir(pathlib.Path(charm.__file__).parent.parent)
    ctx = scenario.Context(charm.HelloWorldCharm)
    container_name = next(iter(ctx.charm_spec.meta["containers"].keys()))
    container = scenario.Container(
        name=container_name,
        can_connect=True,
    )
    state_in = scenario.State(containers={container})
    out = ctx.run(
        ctx.on.pebble_ready(container),
        state_in,
    )
    assert type(out.unit_status) in (scenario.WaitingStatus, scenario.BlockedStatus)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for linters."""

import pathlib
import subprocess
import sys

import pytest

from charmcraft import linters
from charmcraft.models.lint import LintResult


@pytest.mark.slow
@pytest.mark.parametrize(
    "pip_cmd",
    [
        ["--version"],
        ["install", "pytest", "hypothesis"],
    ],
)
def test_pip_check_success(tmp_path: pathlib.Path, pip_cmd: list[str]):
    venv_path = tmp_path / "venv"
    subprocess.run([sys.executable, "-m", "venv", venv_path], check=True)
    subprocess.run([venv_path / "bin" / "python", "-m", "pip", *pip_cmd], check=True)

    lint = linters.PipCheck()
    assert lint.run(tmp_path) == LintResult.OK
    assert lint.text == linters.PipCheck.text


@pytest.mark.slow
@pytest.mark.parametrize(
    "pip_cmd",
    [
        ["install", "--no-deps", "pydantic==2.9.2"],
    ],
)
def test_pip_check_failure(tmp_path: pathlib.Path, pip_cmd: list[str]):
    venv_path = tmp_path / "venv"
    subprocess.run([sys.executable, "-m", "venv", venv_path], check=True)
    subprocess.run([venv_path / "bin" / "python", "-m", "pip", *pip_cmd], check=True)

    lint = linters.PipCheck()
    assert lint.run(tmp_path) == LintResult.WARNING
    assert "pydantic 2.9.2 requires pydantic-core" in lint.text



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for file-related utilities."""

import os
import pathlib
import zipfile

import pytest
from craft_cli import CraftError

from charmcraft.utils.file import build_zip, make_executable, useful_filepath


def test_make_executable_read_bits(tmp_path):
    pth = tmp_path / "test"
    pth.touch(mode=0o640)
    # validity check
    assert pth.stat().st_mode & 0o777 == 0o640
    with pth.open() as fd:
        make_executable(fd)
        # only read bits got made executable
        assert pth.stat().st_mode & 0o777 == 0o750


def test_usefulfilepath_pathlib(tmp_path):
    """Convert the string to Path."""
    test_file = tmp_path / "testfile.bin"
    test_file.touch()
    path = useful_filepath(str(test_file))
    assert path == test_file
    assert isinstance(path, pathlib.Path)


def test_usefulfilepath_home_expanded(tmp_path, monkeypatch):
    """Home-expand the indicated path."""
    fake_home = tmp_path / "homedir"
    fake_home.mkdir()
    test_file = fake_home / "testfile.bin"
    test_file.touch()

    monkeypatch.setitem(os.environ, "HOME", str(fake_home))
    path = useful_filepath("~/testfile.bin")
    assert path == test_file


def test_usefulfilepath_missing():
    """The indicated path is not there."""
    with pytest.raises(CraftError) as cm:
        useful_filepath("not_really_there.txt")
    assert str(cm.value) == "Cannot access 'not_really_there.txt'."


def test_usefulfilepath_inaccessible(tmp_path):
    """The indicated path is not readable."""
    test_file = tmp_path / "testfile.bin"
    test_file.touch(mode=0o000)
    with pytest.raises(CraftError) as cm:
        useful_filepath(str(test_file))
    assert str(cm.value) == f"Cannot access {str(test_file)!r}."


def test_usefulfilepath_not_a_file(tmp_path):
    """The indicated path is not a file."""
    with pytest.raises(CraftError) as cm:
        useful_filepath(str(tmp_path))
    assert str(cm.value) == f"{str(tmp_path)!r} is not a file."


def test_zipbuild_simple(tmp_path):
    """Build a bunch of files in the zip."""
    build_dir = tmp_path / "somedir"
    build_dir.mkdir()

    testfile1 = build_dir / "foo.txt"
    testfile1.write_bytes(b"123\x00456")
    subdir = build_dir / "bar"
    subdir.mkdir()
    testfile2 = subdir / "baz.txt"
    testfile2.write_bytes(b"mo\xc3\xb1o")

    zip_filepath = tmp_path / "testresult.zip"
    build_zip(zip_filepath, build_dir)

    zf = zipfile.ZipFile(zip_filepath)
    assert sorted(x.filename for x in zf.infolist()) == ["bar/baz.txt", "foo.txt"]
    assert zf.read("foo.txt") == b"123\x00456"
    assert zf.read("bar/baz.txt") == b"mo\xc3\xb1o"


def test_zipbuild_symlinks(tmp_path: pathlib.Path):
    """Symlinks are supported."""
    build_dir = tmp_path / "somedir"
    build_dir.mkdir()

    outside_dir = tmp_path / "another_dir"
    outside_dir.mkdir()
    outside_file = outside_dir / "some_file"
    outside_file.write_bytes(b"123\x00456")

    internal_dir = build_dir / "subdirectory"
    internal_dir.mkdir()
    real_file = internal_dir / "real.txt"
    real_file.write_bytes(b"123\x00456")

    internal_file_link = build_dir / "link.txt"
    internal_file_link.symlink_to(real_file)

    internal_dir_link = build_dir / "link_dir"
    internal_dir_link.symlink_to(internal_dir)

    external_file_link = build_dir / "external_link.txt"
    external_file_link.symlink_to(outside_file)

    external_dir_link = build_dir / "external_link_dir"
    external_dir_link.symlink_to(outside_dir)

    zip_filepath = tmp_path / "testresult.zip"
    build_zip(zip_filepath, build_dir)

    zf = zipfile.ZipFile(zip_filepath)

    expected_files = [
        "external_link.txt",
        "external_link_dir/some_file",
        "link.txt",
        "link_dir/real.txt",
        "subdirectory/real.txt",
    ]

    assert sorted(x.filename for x in zf.infolist()) == expected_files
    for file_name in expected_files:
        assert zf.read(file_name) == b"123\x00456"



# Copyright 2021-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Analyze and lint charm structures and files."""

import abc
import ast
import os
import pathlib
import re
import shlex
import subprocess
import sys
import typing
from collections.abc import Collection, Generator
from typing import final

import yaml
from craft_cli import emit

from charmcraft import const, utils
from charmcraft.models.lint import CheckResult, CheckType, LintResult
from charmcraft.models.metadata import CharmMetadataLegacy

# the documentation page for "Analyzers and linters"
BASE_DOCS_URL = "https://juju.is/docs/sdk/charmcraft-analyzers-and-linters"
PYTHON_NAME_REGEX = re.compile(
    r"^([A-Z0-9]([A-Z0-9._-]*[A-Z0-9])?)", flags=re.IGNORECASE
)
MIN_VERSION_REGEX = re.compile(r">=\s*([\d.]+)")
APPROX_VERSION_REGEX = re.compile(r"(?:~=\s*([\d.]+)\.\d|==\s*([\d.]+)\.\*)")
EXACT_VERSION_REGEX = re.compile(r"==\s*([\d.]+)")


def get_entrypoint_from_dispatch(basedir: pathlib.Path) -> pathlib.Path | None:
    """Verify if the charm has a dispatch file pointing to a Python entrypoint.

    :returns: the entrypoint path if all succeeds, None otherwise.
    """
    # get the entrypoint from the last useful dispatch line
    dispatch = basedir / const.DISPATCH_FILENAME
    entrypoint_str = ""
    try:
        with dispatch.open("rt", encoding="utf8") as fh:
            last_line = None
            for line in fh:
                if line.strip():
                    last_line = line
            if last_line:
                entrypoint_str = shlex.split(last_line)[-1]
    except (OSError, UnicodeDecodeError):
        return None
    if not entrypoint_str:
        return None
    return basedir / entrypoint_str


def check_dispatch_with_python_entrypoint(basedir: pathlib.Path) -> pathlib.Path | None:
    """Verify if the charm has a dispatch file pointing to a Python entrypoint.

    :returns: the entrypoint path if all succeeds, None otherwise.
    """
    entrypoint = get_entrypoint_from_dispatch(basedir)
    if entrypoint and entrypoint.suffix == ".py" and os.access(entrypoint, os.X_OK):
        return entrypoint
    return None


class BaseChecker(metaclass=abc.ABCMeta):
    """Base class for checker classes."""

    check_type: CheckType
    name: str
    url: str
    text: str

    exception_result: str

    @abc.abstractmethod
    def run(self, basedir: pathlib.Path) -> str:
        """Run this checker."""
        ...

    @final
    def get_result(self, base_dir: pathlib.Path) -> CheckResult:
        """Get the result of a single checker."""
        try:
            result = self.run(base_dir)
        except Exception as exc:
            result = self.exception_result
            if not self.text:
                self.text = str(exc)
        return CheckResult(
            check_type=self.check_type,
            name=self.name,
            url=self.url,
            text=self.text,
            result=result,
        )

    @final
    def get_ignore_result(self) -> CheckResult:
        """Get the result presuming the checker is ignored."""
        return CheckResult(
            check_type=self.check_type,
            name=self.name,
            url=self.url,
            text="",
            result=LintResult.IGNORED,
        )


class AttributeChecker(BaseChecker, metaclass=abc.ABCMeta):
    """Base attribute checker."""

    check_type = CheckType.ATTRIBUTE
    exception_result = LintResult.UNKNOWN


class Linter(BaseChecker, metaclass=abc.ABCMeta):
    """Base linter class."""

    check_type = CheckType.LINT
    exception_result = LintResult.FATAL
    Result = LintResult


class Language(AttributeChecker):
    """Check the language used to write the charm.

    Currently only Python is detected, if the following checks are true:

    - the charm has a text dispatch with a python call
    - the charm has a `.py` entry point
    - the entry point file is executable
    """

    name = "language"
    url = BASE_DOCS_URL + "#heading--language"
    text = "The charm is written with Python."

    class Result:
        """Possible results for this attribute checker."""

        UNKNOWN = LintResult.UNKNOWN
        PYTHON = "python"

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        python_entrypoint = check_dispatch_with_python_entrypoint(basedir)
        if python_entrypoint is None:
            self.text = "Charm language unknown"
            return self.Result.UNKNOWN
        return self.Result.PYTHON


class Framework(AttributeChecker):
    """Check the framework the charm is based on.

    Currently it detects if the Operator Framework is used, if...

    - the language attribute is set to python
    - the charm contains venv/ops
    - the charm imports ops in the entry point.

    ...or the Reactive Framework is used, if the charm...

    - has a metadata.yaml with "name" in it
    - has a reactive/<name>.py file that imports "charms.reactive"
    - has a file name that starts with "charms.reactive-" inside the "wheelhouse" directory
    """

    name = "framework"
    url = BASE_DOCS_URL + "#heading--framework"

    class Result:
        """Possible results for this attribute checker."""

        OPERATOR = "operator"
        REACTIVE = "reactive"
        UNKNOWN = LintResult.UNKNOWN

    # different texts to be exposed as `text` (see the property below)
    result_texts = {
        Result.OPERATOR: "The charm is based on the Operator Framework.",
        Result.REACTIVE: "The charm is based on the Reactive Framework.",
        Result.UNKNOWN: "The charm is not based on any known Framework.",
    }

    def __init__(self):
        self.result = None
        self.__text = None

    @property
    def text(self) -> str:
        """Return a text in function of the result state."""
        if self.__text:
            return self.__text
        if self.result is None:
            return ""
        return self.result_texts[self.result]

    @text.setter
    def text(self, value: str) -> None:
        self.__text = value

    def _get_imports(self, filepath: pathlib.Path) -> Generator[list[str], None, None]:
        """Parse a Python filepath and yield its imports.

        If the file does not exist or cannot be parsed, return empty. Otherwise
        return the name for each imported module, split by possible dots.
        """
        if not os.access(filepath, os.R_OK):
            return
        try:
            parsed = ast.parse(filepath.read_bytes())
        except SyntaxError:
            return

        for node in ast.walk(parsed):
            if isinstance(node, ast.Import):
                for name in node.names:
                    yield name.name.split(".")
            elif isinstance(node, ast.ImportFrom):
                yield node.module.split(".")

    def _check_operator(self, basedir: pathlib.Path) -> bool:
        """Detect if the Operator Framework is used."""
        python_entrypoint = check_dispatch_with_python_entrypoint(basedir)
        if python_entrypoint is None:
            return False

        opsdir = basedir / const.VENV_DIRNAME / "ops"
        if not opsdir.exists() or not opsdir.is_dir():
            return False

        for import_parts in self._get_imports(python_entrypoint):
            if import_parts[0] == "ops":
                return True
        return False

    def _check_reactive(self, basedir: pathlib.Path) -> bool:
        """Detect if the Reactive Framework is used."""
        try:
            metadata = CharmMetadataLegacy.from_yaml_file(
                basedir / const.METADATA_FILENAME
            )
        except Exception:
            # file not found, corrupted, or mandatory "name" not present
            return False

        wheelhouse_dir = basedir / "wheelhouse"
        if not wheelhouse_dir.exists():
            return False
        if not any(
            f.name.startswith("charms.reactive-") for f in wheelhouse_dir.iterdir()
        ):
            return False

        module_basename = metadata.name.replace("-", "_")
        entrypoint = basedir / "reactive" / f"{module_basename}.py"
        for import_parts in self._get_imports(entrypoint):
            if import_parts[0] == "charms" and import_parts[1] == "reactive":
                return True
        return False

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        self.result = self.Result.UNKNOWN
        if self._check_operator(basedir):
            self.result = self.Result.OPERATOR
        elif self._check_reactive(basedir):
            self.result = self.Result.REACTIVE
        return self.result


class JujuMetadata(Linter):
    """Check that the metadata.yaml file exists and is valid.

    The charm is considered to have a valid metadata if the following checks are true:

    - the metadata.yaml is present
    - it is a valid YAML file
    - it has at least the following fields: name, summary, and description
    """

    name = "metadata"
    url = BASE_DOCS_URL + "#heading--metadata"

    def __init__(self):
        self.text = ""

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        try:
            with (basedir / const.METADATA_FILENAME).open("rt") as md_file:
                metadata = yaml.safe_load(md_file)
        except yaml.YAMLError:
            self.text = "The metadata.yaml file is not a valid YAML file."
            return self.Result.ERROR
        except Exception:
            self.text = "Cannot read the metadata.yaml file."
            return self.Result.ERROR

        # check required attributes
        missing_fields = {"name", "summary", "description"} - set(metadata)
        if missing_fields:
            missing = utils.humanize_list(missing_fields, "and")
            self.text = f"The metadata.yaml file is missing the following attribute(s): {missing}."
            return self.Result.ERROR

        if "series" in metadata:
            self.text = (
                "The metadata.yaml file contains the deprecated attribute: series."
                "This attribute will be rejected starting in Juju 4.0."
            )
            return self.Result.WARNING

        return self.Result.OK


class JujuActions(Linter):
    """Check that the actions.yaml file is valid YAML if it exists."""

    name = "juju-actions"
    url = BASE_DOCS_URL + "#heading--juju-actions"
    text = "The actions.yaml file is not a valid YAML file."

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        filepath = basedir / const.JUJU_ACTIONS_FILENAME
        if not filepath.exists():
            self.text = ""
            # it's optional
            return self.Result.OK

        try:
            with filepath.open("rt", encoding="utf8") as fh:
                yaml.safe_load(fh)
        except Exception:
            return self.Result.ERROR

        self.text = "Valid actions.yaml file."
        return self.Result.OK


class JujuConfig(Linter):
    """Check that the config.yaml file (if it exists) is valid.

    The file is considered valid if the following checks are true:

    - has an 'options' key
    - it is a dictionary
    - each item inside has the mandatory 'type' key
    """

    name = "juju-config"
    url = BASE_DOCS_URL + "#heading--juju-config"

    def __init__(self):
        self.text = ""

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        filepath = basedir / const.JUJU_CONFIG_FILENAME
        if not filepath.exists():
            # it's optional
            return self.Result.OK

        try:
            with filepath.open("rt", encoding="utf8") as fh:
                content = yaml.safe_load(fh)
        except Exception:
            self.text = "The config.yaml file is not a valid YAML file."
            return self.Result.ERROR

        options = content.get("options")
        if not isinstance(options, dict):
            self.text = "Error in config.yaml: must have an 'options' dictionary."
            return self.Result.ERROR

        for value in options.values():
            if "type" not in value:
                self.text = "Error in config.yaml: items under 'options' must have a 'type' key."
                return self.Result.ERROR

        return self.Result.OK


class NamingConventions(Linter):
    """Check that charm follows naming conventions.

    More information can be found at https://juju.is/docs/sdk/styleguide#heading--naming.
    """

    name = "naming-conventions"
    url = "https://juju.is/docs/sdk/styleguide#heading--naming"

    exception_result = LintResult.WARNING

    def __init__(self):
        self.text = ""

    @staticmethod
    def check_naming_convention(names: typing.Iterable[str], scope: str) -> str | None:
        """Check adherence to naming convention.

        :returns: string with warning if present, otherwise None
        """
        snake_keys = [key for key in names if "_" in key]

        if snake_keys:
            hyphen_keys = [key for key in names if "-" in key]

            if hyphen_keys:
                return (
                    f"Some {scope} ({', '.join(snake_keys)}) are in snake case, "
                    f"while others  ({', '.join(hyphen_keys)}) are with hyphens."
                )
            else:
                return (
                    f"Some {scope} ({', '.join(snake_keys)}) are using "
                    "snake case naming convention."
                )

        return None

    @staticmethod
    def _config_options_check(config_file: pathlib.Path) -> list[str]:
        # This is safe as the compliance with YAML is done in the JujuConfig linter
        warnings = []

        if not config_file.exists():
            return warnings

        with config_file.open("rt", encoding="utf8") as fh:
            options = (
                content.get("options", {}) if (content := yaml.safe_load(fh)) else {}
            )

        if check := NamingConventions.check_naming_convention(
            options.keys(), "config-options"
        ):
            warnings.append(check)

        return warnings

    @staticmethod
    def _actions_check(action_file: pathlib.Path) -> list[str]:
        # This is safe as the compliance with YAML is done in the JujuConfig linter
        warnings = []

        if not action_file.exists():
            return warnings

        # This is safe as the compliance with YAML is done in the JujuConfig linter
        with action_file.open("rt", encoding="utf8") as fh:
            if content := yaml.safe_load(fh):
                actions_names = list(dict(content).keys())
            else:
                actions_names = []

        if check := NamingConventions.check_naming_convention(actions_names, "actions"):
            warnings.append(check)

        actions_params = [
            param
            for action_name in actions_names
            if isinstance(content[action_name], dict)
            for param in content.get(action_name, {}).get("params", [])
        ]

        if check := NamingConventions.check_naming_convention(
            actions_params, "action params"
        ):
            warnings.append(check)

        return warnings

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        # Check naming convention on config options

        warnings = NamingConventions._config_options_check(
            basedir / const.JUJU_CONFIG_FILENAME
        ) + NamingConventions._actions_check(basedir / const.JUJU_ACTIONS_FILENAME)

        if warnings:
            all_warning_string = "\n".join(warnings)
            self.text = f"Naming conventions breaks:\n{all_warning_string}"
            return self.exception_result

        return self.Result.OK


class Entrypoint(Linter):
    """Check the entrypoint is correct.

    It validates that the entrypoint, if used by 'dispatch', ...

    - exists
    - is a file
    - is executable
    """

    name = "entrypoint"
    url = BASE_DOCS_URL + "#heading--entrypoint"

    def __init__(self):
        self.text = ""

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        entrypoint = get_entrypoint_from_dispatch(basedir)
        if entrypoint is None:
            self.text = (
                "Cannot find a proper 'dispatch' script pointing to an entrypoint."
            )
            return self.Result.NONAPPLICABLE

        if not entrypoint.exists():
            self.text = f"Cannot find the entrypoint file: {str(entrypoint)!r}"
            return self.Result.ERROR

        if not entrypoint.is_file():
            self.text = f"The entrypoint is not a file: {str(entrypoint)!r}"
            return self.Result.ERROR

        if not os.access(entrypoint, os.X_OK):
            self.text = f"The entrypoint file is not executable: {str(entrypoint)!r}"
            return self.Result.ERROR

        return self.Result.OK


class OpsMainCall(Linter):
    """Check that the entrypoint contains call to ops.main()."""

    name = "ops-main-call"
    url = f"{BASE_DOCS_URL}#heading--ops-main-call"
    text = ""

    def run(self, basedir: pathlib.Path) -> str:
        """Check preconditions and validate there's an ops.main() call."""
        if Framework().run(basedir) != Framework.Result.OPERATOR:
            self.text = "The charm is not based on the operator framework"
            return self.Result.NONAPPLICABLE

        entrypoint = get_entrypoint_from_dispatch(basedir)
        if entrypoint is None:
            self.text = (
                "Cannot find a proper 'dispatch' script pointing to an entrypoint."
            )
            return self.Result.NONAPPLICABLE

        if not entrypoint.exists():
            self.text = f"Cannot find the entrypoint file: {str(entrypoint)!r}"
            return self.Result.NONAPPLICABLE

        if not self._check_main_calls(entrypoint.read_text()):
            self.text = f"The ops.main() call missing from {str(entrypoint)!r}."
            return self.Result.ERROR

        return self.Result.OK

    def _check_main_calls(self, code: str):
        tree = ast.parse(code)
        imports = self._ops_main_imports(tree)
        return self._detect_main_calls(tree, imports=imports)

    def _ops_main_imports(self, tree: ast.AST) -> dict[str, str]:
        """Analyze imports and return a mapping {local_name: imported thing}."""
        rv = {}

        class ImportVisitor(ast.NodeVisitor):
            def visit_Import(self, node: ast.Import):
                for alias in node.names:
                    # Detect statements like `import ops`
                    if alias.name == "ops":
                        rv[alias.asname or alias.name] = "ops"
                    if alias.name == "ops.main" and alias.asname:
                        rv[alias.asname] = "ops.main"
                    elif alias.name.startswith("ops.") and not alias.asname:
                        rv["ops"] = "ops"

            def visit_ImportFrom(self, node: ast.ImportFrom):
                for alias in node.names:
                    # Detect statements like `from ops import main [as ops_main]`
                    if node.module in ("ops", "ops.main") and alias.name == "main":
                        rv[alias.asname or alias.name] = f"{node.module}.main"

        ImportVisitor().visit(tree)
        return rv

    def _detect_main_calls(self, tree: ast.AST, *, imports: dict[str, str]) -> bool:
        main_call_sites = []

        class OpsMainFinder(ast.NodeVisitor):
            def visit_Call(self, node: ast.Call):
                match node.func:
                    # Matches statements like `ops.main.main(...)`
                    case ast.Attribute(
                        value=ast.Attribute(value=ast.Name(id=first), attr=second),
                        attr=third,
                    ):
                        call_site = f"{first}.{second}.{third}(...)"
                    # Matches statements like `ops.main(...)`
                    case ast.Attribute(value=ast.Name(id=first), attr=second):
                        call_site = f"{first}.{second}(...)"
                    # Matches statements like `main(...)`
                    case ast.Name(id=first):
                        call_site = f"{first}(...)"
                    case _:
                        call_site = "_dummy()"

                match = re.match(r"^([a-zA-Z_][a-zA-Z0-9_]*)(.*)", call_site)
                if not match:
                    raise ValueError("impossible")
                alias, rest = match.groups()
                resolved = f"{imports.get(alias, '_dummy')}{rest}"

                if resolved in ("ops.main(...)", "ops.main.main(...)"):
                    main_call_sites.append(call_site)

                self.generic_visit(node)

        OpsMainFinder().visit(tree)
        return any(main_call_sites)


class AdditionalFiles(Linter):
    """Check that the charm does not contain any additional files in the prime directory.

    A few generated files and basic charm files are ignored.
    """

    name = "additional-files"
    text = "No additional files found in the charm."
    url = "https://juju.is/docs/sdk/include-extra-files-in-a-charm"

    IGNORE_FILES: set[pathlib.Path] = {
        pathlib.Path(f)
        for f in (
            {const.CHARMCRAFT_FILENAME, const.MANIFEST_FILENAME}
            | const.CHARM_MANDATORY_FILES
            | const.CHARM_OPTIONAL_FILES
        )
    }

    def _check_additional_files(
        self, stage_dir: pathlib.Path, prime_dir: pathlib.Path
    ) -> str:
        """Compare the staged files with the prime files."""
        errors: list[str] = []
        stage_dir = stage_dir.absolute()
        prime_dir = prime_dir.absolute()

        stage_files = {f.relative_to(stage_dir) for f in stage_dir.rglob("*")}
        prime_files = {f.relative_to(prime_dir) for f in prime_dir.rglob("*")}

        prime_files = prime_files - self.IGNORE_FILES

        for prime_file in prime_files:
            if prime_file not in stage_files:
                errors.append(f"File '{prime_file}' is not staged but in the charm.")

        if errors:
            self.text = "Error: Additional files found in the charm:\n" + "\n".join(
                errors
            )
            return self.Result.ERROR

        return self.Result.OK

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        stage_dir = basedir.parent / "stage"
        if not stage_dir.exists() or not stage_dir.is_dir():
            # Does not work without the build environment
            self.text = (
                "Additional files check not applicable without a build environment."
            )
            return self.Result.NONAPPLICABLE

        return self._check_additional_files(stage_dir, basedir)


class PipCheck(Linter):
    """Check that the pip virtual environment is valid."""

    name = "pip-check"
    text = "Virtual environment is valid."
    url = "https://pip.pypa.io/en/stable/cli/pip_check/"

    def run(self, basedir: pathlib.Path) -> str:
        """Run pip check."""
        venv_dir = basedir / "venv"
        if not venv_dir.is_dir():
            self.text = "Charm does not contain a Python venv."
            return self.Result.NONAPPLICABLE
        if not (venv_dir / "lib").is_dir():
            self.text = "Python venv is not valid."
            return self.Result.NONAPPLICABLE
        python_exe = venv_dir / "bin" / "python"
        delete_parent = False
        if not python_exe.parent.exists():
            delete_parent = True
            python_exe.parent.mkdir()
        if not python_exe.exists():
            delete_python_exe = True
            python_exe.symlink_to(sys.executable)
        else:
            delete_python_exe = False

        pip_cmd = [sys.executable, "-m", "pip", "--python", str(python_exe), "check"]
        try:
            check = subprocess.run(
                pip_cmd,
                text=True,
                capture_output=True,
                check=False,
            )
            if check.returncode == os.EX_OK:
                result = self.Result.OK
            else:
                self.text = check.stdout
                result = self.Result.WARNING
        except (FileNotFoundError, PermissionError) as e:
            self.text = (
                f"{e.strerror}: Could not run Python executable at {sys.executable}."
            )
            result = self.Result.NONAPPLICABLE
        finally:
            if delete_python_exe:
                python_exe.unlink()
            if delete_parent:
                python_exe.parent.rmdir()

        return result


class PyDeps(Linter):
    """Check that all pydeps from all libs are available."""

    name = "pydeps"
    text = "All charmlibs dependencies are included"
    url = "https://documentation.ubuntu.com/charmcraft/stable/howto/manage-libraries/"

    @staticmethod
    def convert_to_fs(name: str) -> str:
        """Convert a package name into its on-disk form.

        This method will match the appearance a module's name would have in
        a site-packages directory.

        This turns "craft-cli", "craft_cli", "Craft-Cli" and "craft.cli" all into "craft_cli".
        """
        return re.sub(r"[-_.]+", "_", name).lower()

    @staticmethod
    def get_version_tuple(version_str: str) -> tuple[int | str, ...]:
        """Get a version tuple from a version string."""

        def coerce_to_int(s: str) -> str | int:
            try:
                return int(s)
            except ValueError:
                return s

        return tuple(coerce_to_int(s) for s in version_str.split("."))

    @classmethod
    def version_matches(
        cls, dep_spec: str, version: tuple[int | str, ...]
    ) -> bool | None:
        """Check if the given version matches the dependency specifier."""
        spec_matched = False
        for match in MIN_VERSION_REGEX.finditer(dep_spec):
            spec_matched = True
            if version >= cls.get_version_tuple(match.group(1)):
                return True
        for match in EXACT_VERSION_REGEX.finditer(dep_spec):
            spec_matched = True
            if version == cls.get_version_tuple(match.group(1)):
                return True
        for match in APPROX_VERSION_REGEX.finditer(dep_spec):
            spec_matched = True
            versions = set(match.group(1, 2)) - {None}
            match_version = cls.get_version_tuple(versions.pop())
            if version[: len(match_version)] == match_version:
                return True

        return not spec_matched

    @classmethod
    def _get_missing_deps(
        cls, deps: Collection[str], venv_path: pathlib.Path
    ) -> tuple[set[str], set[str]]:
        """Get the missing dependencies for a charmlib."""
        libs_path = next((venv_path / "lib").glob("python*")) / "site-packages"

        missing = set()
        non_matching_version = set()
        for dep in deps:
            match = PYTHON_NAME_REGEX.match(dep)
            if not match:
                continue
            name = cls.convert_to_fs(match.group(1))

            # If the package is installed as a distribution, we'll check the version.
            if infos := set(libs_path.glob(f"{name}*.dist-info")):
                info = infos.pop()
                # Trim the ".dist-info" off of the directory name, then extract
                # everything past the last "-" as that will always be the version.
                version_str = info.name[:-10].rpartition("-")[2]
                version = cls.get_version_tuple(version_str)
                if cls.version_matches(dep, version):
                    continue
                non_matching_version.add(dep)
                continue
            # Check if dependency exists as a top-level module
            if (libs_path / name).is_dir() or (libs_path / f"{name}.py").is_file():
                continue
            missing.add(match.group(1))

        return missing, non_matching_version

    def run(self, basedir: pathlib.Path) -> str:
        """Run the pydeps checker."""
        venv_dir = basedir / "venv"
        if not venv_dir.is_dir():
            self.text = "Charm does not contain a Python venv."
            return self.Result.NONAPPLICABLE
        lib_dir = basedir / "lib/charms"
        if not lib_dir.is_dir():
            self.text = "Charm does not have any charmlibs."
            return self.Result.NONAPPLICABLE

        deps = utils.collect_charmlib_pydeps(basedir=basedir)
        try:
            missing_deps, non_matching_versions = self._get_missing_deps(deps, venv_dir)
        except Exception as exc:
            self.text = str(exc)
            emit.debug(exc)
            return self.Result.UNKNOWN

        if missing_deps:
            missing_deps_str = ", ".join(sorted(missing_deps))
            self.text = f"Missing charmlibs dependencies: {missing_deps_str}"
            return self.Result.ERROR

        if non_matching_versions:
            non_matching_versions_str = "\n".join(sorted(non_matching_versions))
            self.text = f"Installed packages don't match charmlibs PYDEPS:\n{non_matching_versions_str}"
            return self.Result.WARNING

        return self.Result.OK


# all checkers to run; the order here is important, as some checkers depend on the
# results from others
CHECKERS: list[type[BaseChecker]] = [
    Language,
    JujuActions,
    JujuConfig,
    JujuMetadata,
    NamingConventions,
    Framework,
    Entrypoint,
    OpsMainCall,
    AdditionalFiles,
    PipCheck,
    PyDeps,
]



from charms.reactive import set_flag, when_not


@when_not("reactivecharm.installed")
def install_reactivecharm():
    # Do your setup here.
    #
    # If your charm has other dependencies before it can install,
    # add those as @when() clauses above., or as additional @when()
    # decorated handlers below
    #
    # See the following for information about reactive charms:
    #
    #  * https://jujucharms.com/docs/devel/developer-getting-started
    #  * https://github.com/juju-solutions/layer-basic#overview
    #
    set_flag("reactivecharm.installed")



# Flask app with a greeting configuration

import flask

app = flask.Flask(__name__)
app.config.from_prefixed_env()


@app.route("/")
def index():
    greeting = app.config.get("GREETING", "Hello, world!")
    return f"{greeting}\n"


if __name__ == "__main__":
    app.run()



# Copyright 2024-2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for packing."""

import pathlib
import zipfile

import craft_platforms
import pytest
import yaml
from craft_application import ServiceFactory
from craft_cli.pytest_plugin import RecordingEmitter

from charmcraft import utils
from charmcraft.application.main import Charmcraft

CURRENT_PLATFORM = utils.get_os_platform()


@pytest.mark.slow
@pytest.mark.skipif(
    CURRENT_PLATFORM.system != "ubuntu",
    reason="Basic charm tests use destructive mode.",
)
def test_build_basic_charm(
    monkeypatch: pytest.MonkeyPatch,
    emitter: RecordingEmitter,
    new_path: pathlib.Path,
    service_factory: ServiceFactory,
    app: Charmcraft,
):
    monkeypatch.setenv("CRAFT_DEBUG", "1")
    monkeypatch.setattr(
        "sys.argv",
        ["charmcraft", "pack", "--destructive-mode"],
    )
    (new_path / "requirements.txt").write_text(
        "distro==1.4.0"
    )  # Any requirement will do.

    app.configure({})
    assert app.run() == 0

    charm_files = list(new_path.glob("example-charm_*.charm"))
    assert len(charm_files) == 1

    with zipfile.ZipFile(charm_files[0]) as charm_zip:
        metadata = yaml.safe_load(charm_zip.read("metadata.yaml"))
        manifest = yaml.safe_load(charm_zip.read("manifest.yaml"))

    emitter.assert_progress(f"Packing charm {charm_files[0].name}")

    assert "bases" in manifest
    project = app.services.get("project").get().marshal()
    if "platforms" in project:
        base_name = manifest["bases"][0]["name"]
        base_version = manifest["bases"][0]["channel"]
        assert f"{base_name}@{base_version}" == project["base"]
    else:
        assert manifest["bases"][0]["name"] == CURRENT_PLATFORM.system
        assert manifest["bases"][0]["channel"] == CURRENT_PLATFORM.release
    assert (
        manifest["bases"][0]["architectures"][0]
        == craft_platforms.DebianArchitecture.from_host()
    )

    assert metadata["name"] == project["name"]
    assert metadata["summary"] == project["summary"]
    assert metadata["description"] == project["description"]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for init command."""

import argparse
import contextlib
import os
import pathlib
import re
import shutil
import subprocess
import sys
from unittest import mock

import pytest
import pytest_check

import charmcraft
from charmcraft import errors
from charmcraft.application.commands import init
from charmcraft.utils import S_IXALL

with contextlib.suppress(ImportError):
    import pwd

BASIC_INIT_FILES = frozenset(
    pathlib.Path(p)
    for p in (
        ".gitignore",
        "charmcraft.yaml",
        "CONTRIBUTING.md",
        "LICENSE",
        "pyproject.toml",
        "README.md",
        "src",
        "src/charm.py",
        "tests",
        "tests/integration",
        "tests/integration/conftest.py",
        "tests/integration/test_charm.py",
        "tests/unit",
        "tests/unit/test_charm.py",
        "tox.ini",
        "uv.lock",
    )
)
UNKNOWN_AUTHOR_REGEX = re.compile(
    r"^Unable to automatically determine author's name, specify it with --author$"
)
BAD_CHARM_NAME_REGEX = re.compile(
    r" is not a valid charm name. The name must start with a lowercase letter and contain only alphanumeric characters and hyphens.$",
)
VALID_AUTHORS = [
    pytest.param("Author McAuthorFace", id="ascii-author"),
    pytest.param("فلانة الفلانية", id="non-ascii-author"),
]


@pytest.fixture
def init_command():
    return init.InitCommand(
        {"app": charmcraft.application.APP_METADATA, "services": None}
    )


def create_namespace(
    *,
    name="my-charm",
    author="J Doe",
    force=False,
    profile=init.DEFAULT_PROFILE,
    project_dir: pathlib.Path | None = None,
):
    """Helper to create a valid namespace."""
    if project_dir is None:
        project_dir = pathlib.Path.cwd()
    return argparse.Namespace(
        name=name,
        author=author,
        force=force,
        profile=profile,
        project_dir=project_dir,
    )


@pytest.mark.parametrize(
    ("profile", "base_expected_files", "has_workload_module"),
    [
        pytest.param("machine", BASIC_INIT_FILES, True, id="machine"),
        pytest.param("kubernetes", BASIC_INIT_FILES, True, id="kubernetes"),
    ],
)
@pytest.mark.parametrize(
    ("charm_name", "workload_module_filename"),
    [
        pytest.param("machine", "workload.py", id="generic_name"),
        pytest.param(
            "foo-bar-k8s-operator", "foo_bar.py", id="generic_suffix_k8s_operator"
        ),
        pytest.param("foo-bar-k8s", "foo_bar.py", id="generic_suffix_k8s"),
        pytest.param("foo-bar-operator", "foo_bar.py", id="generic_suffix_operator"),
        pytest.param("charm123", "charm123.py", id="name_with_numbers"),
    ],
)
@pytest.mark.parametrize("author", VALID_AUTHORS)
def test_files_created_correct(
    new_path,
    init_command,
    profile: str,
    base_expected_files: set[pathlib.Path],
    has_workload_module: bool,
    charm_name,
    workload_module_filename: str,
    author,
):
    params = create_namespace(name=charm_name, author=author, profile=profile)
    init_command.run(params)

    if has_workload_module:
        workload_module_path = pathlib.Path(f"src/{workload_module_filename}")
        expected_files = base_expected_files.union({workload_module_path})
    else:
        expected_files = base_expected_files

    actual_files = {p.relative_to(new_path) for p in new_path.rglob("*")}

    charmcraft_yaml = (new_path / "charmcraft.yaml").read_text()
    tox_ini = (new_path / "tox.ini").read_text(encoding="utf-8")

    pytest_check.equal(actual_files, expected_files)
    pytest_check.is_true(
        re.search(rf"^name: {charm_name}$", charmcraft_yaml, re.MULTILINE)
    )
    pytest_check.is_true(re.search(rf"^# Copyright \d+ {author}", tox_ini))


def test_force(new_path, init_command):
    tmp_file = new_path / "README.md"
    with tmp_file.open("w") as f:
        f.write("This is a nonsense readme")

    init_command.run(create_namespace(force=True))

    # Check that init ran
    assert (new_path / "LICENSE").exists()

    # Check that init did not overwrite files
    with tmp_file.open("r") as f:
        assert f.read() == "This is a nonsense readme"


@pytest.mark.parametrize("name", [None, 0, "1234", "yolo swag", "camelCase"])
def test_bad_name(monkeypatch, new_path, init_command, name):
    with pytest.raises(errors.CraftError, match=BAD_CHARM_NAME_REGEX):
        init_command.run(create_namespace(name=name))


@pytest.mark.skipif(sys.platform == "win32", reason=("Password database only on Unix"))
@pytest.mark.parametrize("author", VALID_AUTHORS)
def test_gecos_valid_author(monkeypatch, new_path, init_command, author):
    monkeypatch.setattr(
        pwd,
        "getpwuid",
        mock.Mock(
            return_value=pwd.struct_passwd(
                ("user", "pass", 1, 1, f"{author},,,", "homedir", "shell")
            )
        ),
    )

    init_command.run(create_namespace(author=None))

    pytest_check.is_true(
        re.search(rf"^# Copyright \d+ {author}", (new_path / "tox.ini").read_text())
    )


@pytest.mark.skipif(sys.platform == "win32", reason=("Password database only on Unix"))
@pytest.mark.parametrize(
    ("mock_getpwuid", "error_msg"),
    [
        pytest.param(
            mock.Mock(side_effect=KeyError("no user")),
            UNKNOWN_AUTHOR_REGEX,
            id="user-doesnt-exist",
        ),
    ],
)
def test_gecos_user_not_found(
    monkeypatch, new_path, init_command, mock_getpwuid, error_msg
):
    monkeypatch.setattr(pwd, "getpwuid", mock_getpwuid)

    with pytest.raises(errors.CraftError, match=error_msg):
        init_command.run(create_namespace(author=None))


@pytest.mark.skipif(sys.platform == "win32", reason=("Password database only on Unix"))
def test_gecos_user_has_no_name(monkeypatch, new_path, init_command):
    mock_getpwuid = mock.Mock(
        return_value=pwd.struct_passwd(("user", "pass", 1, 1, "", "dir", "shell"))
    )
    monkeypatch.setattr(pwd, "getpwuid", mock_getpwuid)

    with pytest.raises(errors.CraftError, match=UNKNOWN_AUTHOR_REGEX):
        init_command.run(create_namespace(author=None))


@pytest.mark.parametrize(
    "subdir",
    [
        "somedir",
        "some/dir",
        "a/really/deep/path/with_parents/all/created",
        pytest.param(
            "/tmp/test_charm_dir-absolute_directory_path",
            marks=pytest.mark.skipif(os.name != "posix", reason="This is a posix path"),
        ),
    ],
)
@pytest.mark.parametrize("base_expected_files", [BASIC_INIT_FILES])
def test_create_directory(new_path, init_command, subdir, base_expected_files):
    init_dir = new_path / subdir

    try:
        params = create_namespace(name="foo-bar-k8s", project_dir=pathlib.Path(subdir))
        init_command.run(params)

        workload_module_path = pathlib.Path("src/foo_bar.py")
        expected_files = base_expected_files.union({workload_module_path})

        actual_files = {p.relative_to(init_dir) for p in init_dir.rglob("*")}

        assert actual_files == expected_files

    finally:
        shutil.rmtree(init_dir)


@pytest.mark.skipif(os.name != "posix", reason="Checking posix executable bit.")
def test_executable_set(new_path, init_command):
    init_command.run(create_namespace())

    for path in new_path.rglob(".py"):
        assert path.stat().st_mode & S_IXALL == S_IXALL


@pytest.mark.slow
@pytest.mark.skipif(
    bool(os.getenv("RUNNING_TOX")) and sys.version_info < (3, 11),
    reason="does not work inside tox in Python3.10 and below",
)
@pytest.mark.parametrize("profile", list(init.PROFILES))
def test_tox_success(new_path, init_command, profile):
    # fix the PYTHONPATH and PATH so the tests in the initted environment use our own
    # virtualenv libs and bins (if any), as they need them, but we're not creating a
    # venv for the local tests (note that for CI doesn't use a venv)
    env = os.environ.copy()
    env_paths = [p for p in sys.path if "env/lib/python" in p]
    if env_paths:
        if "PYTHONPATH" in env:
            env["PYTHONPATH"] += ":" + ":".join(env_paths)
        else:
            env["PYTHONPATH"] = ":".join(env_paths)
        for path in env_paths:
            bin_path = path[: path.index("env/lib/python")] + "env/bin"
            env["PATH"] = bin_path + ":" + env["PATH"]

    init_command.run(create_namespace(profile=profile))

    if not (new_path / "tox.ini").exists():
        pytest.skip("init template doesn't contain tox.ini file")

    if list((new_path / "tests").glob("*.py")):  # If any tests exist
        result = subprocess.run(
            ["tox", "-v", "run", "-e", "unit"],
            cwd=new_path,
            env=env,
            text=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=False,
        )
        assert result.returncode == 0, "Tox run failed:\n" + result.stdout



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for charmlibs service."""

import pytest

from charmcraft.services.charmlibs import CharmLibDelta, CharmLibsService
from charmcraft.store.models import Library


@pytest.fixture
def service(service_factory):
    return service_factory.get("charm_libs")


# This test is marked as slow because it hits the real charmhub.
@pytest.mark.slow
@pytest.mark.parametrize(
    ("local_libs", "expected"),
    [
        pytest.param({}, [], id="no-libs"),
        pytest.param(
            {
                "lib/charms/example_charm/v0/test_lib.py": 'LIBID = "e000776021fd4b73ade744727654ac72"\nLIBAPI = 0\nLIBPATCH = 1'
            },
            [],
            id="up-to-date-lib",
        ),
        pytest.param(
            {
                "lib/charms/example_charm/v0/unpublished_lib.py": 'LIBID = "unpublished"\nLIBAPI = 0\nLIBPATCH = 1'
            },
            [CharmLibDelta("unpublished_lib", (0, 1), None)],
            id="unpublished-lib",
        ),
        pytest.param(
            {
                "lib/charms/example_charm/v0/test_lib.py": 'LIBID = "e000776021fd4b73ade744727654ac72"\nLIBAPI = 0\nLIBPATCH = 2'
            },
            [CharmLibDelta("test_lib", (0, 2), (0, 1))],
            id="out-of-date-lib",
        ),
    ],
)
def test_get_unpublished_libs(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
    service: CharmLibsService,
    local_libs: dict[str, Library],
    expected: list[CharmLibDelta],
):
    service._project_dir = tmp_path
    for lib_path_str, lib_contents in local_libs.items():
        lib_path = tmp_path / lib_path_str
        lib_path.parent.mkdir(parents=True)
        lib_path.write_text(lib_contents)

    assert service.get_unpublished_libs() == expected



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for charmlibs service."""

import pathlib

import pytest
import pytest_mock

from charmcraft import utils
from charmcraft.services.charmlibs import CharmLibDelta, CharmLibsService
from charmcraft.store.models import Library


@pytest.fixture
def service(service_factory):
    return service_factory.get("charm_libs")


@pytest.fixture(params=["my-charm", "your-charm"])
def charm_name(request) -> str:
    return request.param


@pytest.fixture(params=["my_lib", "your_lib"])
def lib_name(request) -> str:
    return request.param


@pytest.fixture(params=[0, 1])
def api(request) -> int:
    return request.param


@pytest.fixture(params=[None, 0, 1])
def patch(request) -> int | None:
    return request.param


def test_is_downloaded_no_file(
    fake_project_dir: pathlib.Path,
    service: CharmLibsService,
    charm_name: str,
    lib_name: str,
    api: int,
    patch: int | None,
):
    assert not service.is_downloaded(
        charm_name=charm_name, lib_name=lib_name, api=api, patch=patch
    )


@pytest.mark.parametrize(("patch", "expected"), [(None, True), (1, True), (2, False)])
def test_is_downloaded_with_file(
    project_path: pathlib.Path,
    service: CharmLibsService,
    charm_name: str,
    lib_name: str,
    patch: int | None,
    expected: bool,
):
    lib_path = project_path / utils.get_lib_path(charm_name, lib_name, 0)
    lib_path.parent.mkdir(parents=True)
    lib_path.write_text("LIBID='abc'\nLIBAPI=0\nLIBPATCH=1\n")

    assert (
        service.is_downloaded(
            charm_name=charm_name, lib_name=lib_name, api=0, patch=patch
        )
        == expected
    )


@pytest.mark.parametrize(
    ("charm_name", "lib_name", "lib_contents", "expected"),
    [
        pytest.param(
            "my-charm",
            "my_lib",
            "LIBID='abc'\nLIBAPI=0\nLIBPATCH=1\n",
            (0, 1),
            id="0.1",
        ),
        pytest.param(
            "my-charm",
            "my_lib",
            "LIBID='abc'\nLIBAPI=16\nLIBPATCH=19\n",
            (16, 19),
            id="16.19",
        ),
        pytest.param(
            "my-charm",
            "my_lib",
            "LIBID='abc'\nLIBAPI=0\nLIBPATCH=-1\n",
            None,
            id="patch_negative_1",
        ),
        pytest.param("my-charm", "my_lib", None, None, id="nonexistent"),
    ],
)
def test_get_local_version(
    project_path: pathlib.Path,
    service: CharmLibsService,
    charm_name: str,
    lib_name: str,
    lib_contents: str | None,
    expected: tuple[int, int] | None,
):
    if expected is not None:
        lib_path = project_path / utils.get_lib_path(charm_name, lib_name, expected[0])
        (project_path / lib_path).parent.mkdir(parents=True)
        (project_path / lib_path).write_text(str(lib_contents))

    assert (
        service.get_local_version(charm_name=charm_name, lib_name=lib_name) == expected
    )


@pytest.mark.parametrize(
    "lib",
    [
        Library("lib_id", "lib_name", "charm_name", 0, 0, "some content", "hashy"),
    ],
)
def test_write_success(
    project_path: pathlib.Path, service: CharmLibsService, lib: Library
):
    service.write(lib)

    actual = (
        project_path / utils.get_lib_path(lib.charm_name, lib.lib_name, lib.api)
    ).read_text()

    assert actual == lib.content


@pytest.mark.parametrize(
    "lib",
    [
        Library("lib_id", "lib_name", "charm_name", 0, 0, None, "hashy"),
    ],
)
def test_write_error(
    fake_project_dir: pathlib.Path, service: CharmLibsService, lib: Library
):
    with pytest.raises(ValueError, match="Library has no content"):
        service.write(lib)


@pytest.mark.parametrize(
    ("store_libs", "expected"),
    [
        pytest.param(
            {
                "abc": Library(
                    "abc",
                    "test-lib",
                    "charmcraft-test-charm",
                    0,
                    1000,
                    content=None,
                    content_hash="hash",
                )
            },
            [],
            id="all-up-to-date",
        ),
        pytest.param(
            {}, [CharmLibDelta("test_lib", (0, 1000), None)], id="all-unpublished"
        ),
        pytest.param(
            {
                "abc": Library(
                    "abc",
                    "test-lib",
                    "charmcraft-test-charm",
                    0,
                    999,
                    content=None,
                    content_hash="hash",
                )
            },
            [CharmLibDelta("test_lib", (0, 1000), (0, 999))],
            id="out-of-date",
        ),
    ],
)
def test_get_unpublished_libs(
    fake_project_dir: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
    service: CharmLibsService,
    mocker: pytest_mock.MockerFixture,
    store_libs: dict[str, Library],
    expected: list[CharmLibDelta],
):
    service._project_dir = fake_project_dir
    local_lib_dir = fake_project_dir / "lib/charms/example_charm/v0"
    local_lib_dir.mkdir(parents=True)
    local_lib_file = local_lib_dir / "test_lib.py"
    local_lib_file.write_text("LIBID='abc'\nLIBAPI=0\nLIBPATCH=1000")

    mocker.patch.object(
        service._services.get("store"),
        "get_libraries_metadata_by_id",
        return_value=store_libs,
    )

    assert service.get_unpublished_libs() == expected



import pytest
from ops import testing

from charm import TestCmdCharm


def mock_get_version():
    """Get a mock version string without executing the workload code."""
    return "1.0.0"


def test_start(monkeypatch: pytest.MonkeyPatch):
    """Test that the charm has the correct state after handling the start event."""
    # Arrange:
    ctx = testing.Context(TestCmdCharm)
    monkeypatch.setattr("charm.test_cmd.get_version", mock_get_version)
    # Act:
    state_out = ctx.run(ctx.on.start(), testing.State())
    # Assert:
    assert state_out.workload_version is not None
    assert state_out.unit_status == testing.ActiveStatus()



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for the provider service."""

import functools
import pathlib
from collections.abc import Iterator
from unittest import mock

import craft_application
import pytest
from craft_cli.pytest_plugin import RecordingEmitter
from craft_providers import bases

from charmcraft.application.main import APP_METADATA
from charmcraft.services.provider import ProviderService, _maybe_lock_cache


@pytest.fixture
def provider_service(
    monkeypatch: pytest.MonkeyPatch,
    fake_path: pathlib.Path,
    service_factory: craft_application.ServiceFactory,
) -> craft_application.ProviderService:
    # Workaround for https://github.com/canonical/craft-application/issues/816
    monkeypatch.delenv("SNAP", raising=False)
    fake_cache_dir = fake_path / "cache"
    fake_cache_dir.mkdir(parents=True)

    service_factory.update_kwargs(
        "provider",
        work_dir=fake_path,
        provider_name="host",
    )

    return service_factory.get("provider")


@pytest.fixture
def mock_register(monkeypatch) -> Iterator[mock.Mock]:
    register = mock.Mock()
    monkeypatch.setattr("atexit.register", register)
    yield register

    # Call the exit hooks as if exiting the application.
    for hook in register.mock_calls:
        functools.partial(*hook.args)()


@pytest.mark.parametrize(
    "base_name",
    [
        bases.BaseName("ubuntu", "20.04"),
        bases.BaseName("ubuntu", "22.04"),
        bases.BaseName("ubuntu", "24.04"),
        bases.BaseName("ubuntu", "devel"),
        bases.BaseName("almalinux", "9"),
    ],
)
def test_get_base_forwards_cache(
    monkeypatch,
    provider_service: ProviderService,
    fake_path: pathlib.Path,
    base_name: bases.BaseName,
):
    monkeypatch.setattr(
        "charmcraft.env.get_host_shared_cache_path", lambda: fake_path / "cache"
    )

    base = provider_service.get_base(
        base_name=base_name,
        instance_name="charmcraft-test-instance",
    )

    assert base._cache_path == fake_path / "cache"


@pytest.mark.parametrize(
    "base_name",
    [
        bases.BaseName("ubuntu", "20.04"),
        bases.BaseName("ubuntu", "22.04"),
        bases.BaseName("ubuntu", "24.04"),
        bases.BaseName("ubuntu", "devel"),
        bases.BaseName("almalinux", "9"),
    ],
)
def test_get_base_no_cache_if_locked(
    monkeypatch,
    mock_register,
    tmp_path: pathlib.Path,
    base_name: bases.BaseName,
    emitter: RecordingEmitter,
):
    cache_path = tmp_path / "cache"
    cache_path.mkdir(exist_ok=True, parents=True)

    # Make a new path object to work around caching the paths and thus getting the
    # same file descriptor.
    locked = _maybe_lock_cache(cache_path)
    assert locked
    new_cache_path = pathlib.Path(str(cache_path))
    monkeypatch.setattr(
        "charmcraft.env.get_host_shared_cache_path", lambda: new_cache_path
    )

    # Can't use the fixture as pyfakefs doesn't handle locks.
    provider_service = ProviderService(
        app=APP_METADATA,
        services=None,  # pyright: ignore[reportArgumentType]
        work_dir=tmp_path,
    )

    base = provider_service.get_base(
        base_name=base_name,
        instance_name="charmcraft-test-instance",
    )

    assert base._cache_path is None
    emitter.assert_progress(
        "Shared cache locked by another process; running without cache.",
        permanent=True,
    )


def test_maybe_lock_cache_locks_single_lock(tmp_path: pathlib.Path) -> None:
    assert _maybe_lock_cache(tmp_path)


def test_maybe_lock_cache_with_another_lock(tmp_path: pathlib.Path) -> None:
    # Need to save the open file so it's not closed when we try a second time.
    first_file_descriptor = _maybe_lock_cache(tmp_path)
    assert first_file_descriptor
    assert _maybe_lock_cache(tmp_path) is None



# Copyright 2023-2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""New entrypoint for charmcraft."""

from __future__ import annotations

import datetime
from typing import Any

import craft_application
import craft_cli
from craft_application import util
from craft_application.errors import ProjectFileMissingError
from craft_parts.plugins.plugins import PluginType
from overrides import override

from charmcraft import const, models, parts, services
from charmcraft.application import commands

GENERAL_SUMMARY = """
Charmcraft helps build, package and publish operators on Charmhub.

Together with the Python Operator Framework, charmcraft simplifies
operator development and collaboration.

See https://charmhub.io/publishing for more information.
"""

APP_METADATA = craft_application.AppMetadata(
    name="charmcraft",
    summary=GENERAL_SUMMARY,
    ProjectClass=models.CharmcraftProject,
    source_ignore_patterns=["*.charm", "charmcraft.yaml"],
    docs_url="https://documentation.ubuntu.com/charmcraft/{version}",
    supports_multi_base=True,
    mandatory_adoptable_fields=[],  # Version field is not mandatory.
)

PRIME_BEHAVIOUR_CHANGE_MESSAGE = (
    "IMPORTANT: The behaviour of the 'prime' keyword has changed in Charmcraft 3. This "
    "keyword will no longer add files that would otherwise be excluded from the "
    "charm, instead filtering existing files. Additional files may be added using the "
    "'dump' plugin.\n"
    "To include extra files, see: https://juju.is/docs/sdk/include-extra-files-in-a-charm"
)


class Charmcraft(craft_application.Application):
    """Charmcraft application definition."""

    def __init__(
        self,
        app: craft_application.AppMetadata,
        services: craft_application.ServiceFactory,
    ) -> None:
        super().__init__(app=app, services=services, extra_loggers={"charmcraft"})
        self._global_args: dict[str, Any] = {}
        self._dispatcher: craft_cli.Dispatcher | None = None
        self._cli_loggers |= {"charmcraft"}

    @property
    def command_groups(self) -> list[craft_cli.CommandGroup]:
        """Return command groups."""
        return self._command_groups

    def _check_deprecated(self, yaml_data: dict[str, Any]) -> None:
        """Check for deprecated fields in the yaml_data."""
        # We only need to warn people once.
        if self.is_managed():
            return
        has_primed_part = False
        if "parts" in yaml_data:
            prime_changed_extensions = {"charm", "reactive"}
            for name, part in yaml_data["parts"].items():
                if not {name, part.get("plugin", None)} & prime_changed_extensions:
                    continue
                if "prime" in part:
                    has_primed_part = True
        if has_primed_part:
            craft_cli.emit.progress(PRIME_BEHAVIOUR_CHANGE_MESSAGE, permanent=True)

    def _configure_services(self, provider_name: str | None) -> None:
        super()._configure_services(provider_name)
        self.services.update_kwargs(
            "charm_libs",
            project_dir=self.project_dir,
        )
        self.services.update_kwargs(
            "provider",
            work_dir=self.project_dir,
        )

    def configure(self, global_args: dict[str, Any]) -> None:
        """Configure the application using any global arguments."""
        super().configure(global_args)
        self._global_args = global_args
        if not util.is_managed_mode():
            self.services.get("state").set(
                "charmcraft",
                "started_at",
                value=datetime.datetime.now().isoformat(),
                overwrite=True,
            )

    def _get_dispatcher(self) -> craft_cli.Dispatcher:
        """Get the dispatcher, with a charmcraft-specific side-effect of storing it on the app."""
        self._dispatcher = super()._get_dispatcher()
        return self._dispatcher

    @override
    def _get_app_plugins(self) -> dict[str, PluginType]:
        plugins = parts.get_app_plugins()

        try:
            full_build_plan = self.services.get("build_plan").create_build_plan(
                platforms=None, build_on=None, build_for=None
            )
        except ProjectFileMissingError:
            return plugins
        bases = {build_info.build_base for build_info in full_build_plan}
        for base in bases:
            if str(base) not in const.CHARM_OR_REACTIVE_BASES:
                plugins.pop("charm")
                plugins.pop("reactive")
                break

        return plugins

    @override
    def _run_inner(self) -> int:
        # We need to overwrite the lookup directory if the user has provided a different project
        # path.
        if not util.is_managed_mode():
            dispatcher = self._get_dispatcher()
            dispatcher.load_command(self.app_config)
            parsed_args = dispatcher.parsed_args()
            self.project_dir = (
                getattr(parsed_args, "project_dir", self.project_dir)
                .expanduser()
                .resolve()
            )
            self.services.update_kwargs(
                "project",
                project_dir=self.project_dir,
            )
        return super()._run_inner()


def create_app() -> Charmcraft:
    """Create the Charmcraft application with its commands."""
    services.register_services()
    charmcraft_services = services.ServiceFactory(app=APP_METADATA)
    app = Charmcraft(app=APP_METADATA, services=charmcraft_services)
    commands.fill_command_groups(app)

    return app


def get_app_info() -> tuple[craft_cli.Dispatcher, dict[str, Any]]:
    """Retrieve application info. Used by craft-cli's completion module."""
    app = create_app()
    dispatcher = app._create_dispatcher()

    return dispatcher, app.app_config


def main() -> int:
    """Run craft-application based charmcraft."""
    app = create_app()

    return app.run()



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Craft-parts plugins and plugin overrides for charmcraft."""

from ._charm import CharmPlugin, CharmPluginProperties
from ._poetry import PoetryPlugin, PoetryPluginProperties
from ._python import PythonPlugin, PythonPluginProperties
from ._reactive import ReactivePlugin, ReactivePluginProperties
from ._uv import UvPlugin
from craft_parts.plugins.uv_plugin import UvPluginProperties

__all__ = [
    "CharmPlugin",
    "CharmPluginProperties",
    "PoetryPlugin",
    "PoetryPluginProperties",
    "PythonPlugin",
    "PythonPluginProperties",
    "ReactivePlugin",
    "ReactivePluginProperties",
    "UvPlugin",
    "UvPluginProperties",
]



#!/usr/bin/env python3
# Copyright 2025 Alex Lowe
# See LICENSE file for licensing details.

"""Charm the application."""

import logging

import ops

logger = logging.getLogger(__name__)


class PlatformsCharm(ops.CharmBase):
    """Charm the application."""

    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        framework.observe(self.on.start, self._on_start)

    def _on_start(self, event: ops.StartEvent):
        """Handle start event."""
        self.unit.status = ops.ActiveStatus()


if __name__ == "__main__":  # pragma: nocover
    ops.main(PlatformsCharm)  # type: ignore



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for dispatch script creation."""

import pathlib

import pytest
import pytest_check

from charmcraft import const, dispatch


def test_create_dispatch_hooks_exist(fake_path: pathlib.Path):
    """Test that nothing happens if a hooks directory exists."""
    prime_dir = fake_path / "prime"
    (prime_dir / const.HOOKS_DIRNAME).mkdir(parents=True)

    pytest_check.is_false(dispatch.create_dispatch(prime_dir=prime_dir))

    pytest_check.is_false((prime_dir / const.DISPATCH_FILENAME).exists())


def test_create_dispatch_dispatch_exists(fake_path: pathlib.Path):
    """Test that nothing happens if dispatch file already exists."""
    prime_dir = fake_path / "prime"
    prime_dir.mkdir()
    dispatch_path = prime_dir / const.DISPATCH_FILENAME
    dispatch_path.write_text("DO NOT OVERWRITE")

    pytest_check.is_false(dispatch.create_dispatch(prime_dir=prime_dir))

    pytest_check.equal(dispatch_path.read_text(), "DO NOT OVERWRITE")


@pytest.mark.parametrize("entrypoint", ["src/charm.py", "src/some_entrypoint.py"])
def test_create_dispatch_no_entrypoint(fake_path: pathlib.Path, entrypoint):
    prime_dir = fake_path / "prime"
    prime_dir.mkdir()
    dispatch_path = prime_dir / const.DISPATCH_FILENAME

    pytest_check.is_false(
        dispatch.create_dispatch(prime_dir=prime_dir, entrypoint=entrypoint)
    )

    pytest_check.is_false(dispatch_path.exists())


@pytest.mark.parametrize("entrypoint", ["src/charm.py", "src/some_entrypoint.py"])
def test_create_dispatch_with_entrypoint(fake_path: pathlib.Path, entrypoint):
    prime_dir = fake_path / "prime"
    prime_dir.mkdir()
    entrypoint = prime_dir / entrypoint
    entrypoint.parent.mkdir(parents=True, exist_ok=True)
    entrypoint.touch()
    dispatch_file = prime_dir / const.DISPATCH_FILENAME
    expected = dispatch.DISPATCH_SCRIPT_TEMPLATE.format(entrypoint=entrypoint)

    pytest_check.is_true(
        dispatch.create_dispatch(prime_dir=prime_dir, entrypoint=entrypoint)
    )
    pytest_check.equal(dispatch_file.read_text(), expected)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for package service."""

import datetime
import pathlib

import freezegun
import pytest
import pytest_check

import charmcraft
from charmcraft import const
from charmcraft.application.main import APP_METADATA
from charmcraft.services.package import PackageService


@pytest.fixture(
    params=[
        pytest.param(path, id=path.name)
        for path in (pathlib.Path(__file__).parent / "sample_projects").iterdir()
    ]
)
def project_path(request: pytest.FixtureRequest) -> pathlib.Path:
    return request.param / "project"


@pytest.fixture
def fake_project_yaml(project_path: pathlib.Path) -> str:
    return (project_path / "charmcraft.yaml").read_text()


@pytest.fixture
def package_service(
    project_path: pathlib.Path,
    service_factory,
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.chdir(project_path)
    svc = PackageService(
        app=APP_METADATA,
        services=service_factory,
    )
    service_factory.package = svc
    return svc


@freezegun.freeze_time(
    datetime.datetime(2020, 3, 14, 0, 0, 0, tzinfo=datetime.timezone.utc)
)
def test_write_metadata(monkeypatch, new_path, package_service, project_path):
    monkeypatch.setattr(charmcraft, "__version__", "3.0-test-version")
    test_prime_dir = new_path / "prime"
    test_prime_dir.mkdir()
    expected_prime_dir = project_path.parent / "prime"

    package_service.write_metadata(test_prime_dir)

    for file in expected_prime_dir.iterdir():
        pytest_check.equal((test_prime_dir / file.name).read_text(), file.read_text())


@freezegun.freeze_time(
    datetime.datetime(2020, 3, 14, 0, 0, 0, tzinfo=datetime.timezone.utc)
)
def test_overwrite_metadata(monkeypatch, new_path, package_service, project_path):
    """Test that the metadata file gets rewritten for a charm.

    Regression test for https://github.com/canonical/charmcraft/issues/1654
    """
    monkeypatch.setattr(charmcraft, "__version__", "3.0-test-version")
    test_prime_dir = new_path / "prime"
    test_prime_dir.mkdir()
    expected_prime_dir = project_path.parent / "prime"

    (test_prime_dir / const.METADATA_FILENAME).write_text("INVALID!!")

    package_service.write_metadata(test_prime_dir)

    for file in expected_prime_dir.iterdir():
        pytest_check.equal((test_prime_dir / file.name).read_text(), file.read_text())


@pytest.mark.parametrize(
    "project_path",
    [pathlib.Path(__file__).parent / "sample_projects" / "basic-reactive" / "project"],
)
@freezegun.freeze_time(
    datetime.datetime(2020, 3, 14, 0, 0, 0, tzinfo=datetime.timezone.utc)
)
def test_no_overwrite_reactive_metadata(monkeypatch, new_path, package_service):
    """Test that the metadata file doesn't get overwritten for a reactive charm..

    Regression test for https://github.com/canonical/charmcraft/issues/1654
    """
    monkeypatch.setattr(charmcraft, "__version__", "3.0-test-version")
    test_prime_dir = new_path / "prime"
    test_prime_dir.mkdir()
    test_stage_dir = new_path / "stage"
    test_stage_dir.mkdir()
    (test_stage_dir / const.METADATA_FILENAME).write_text("INVALID!!")

    package_service.write_metadata(test_prime_dir)

    assert not (test_prime_dir / const.METADATA_FILENAME).exists()



# Copyright 2024 Canonical Ltd.
import ops  # type: ignore


class SomeCharm(ops.CharmBase): ...


# ruff: noqa: ERA001
# charmcraft analyse should detect that ops.main() call is missing
#
# if __name__ == "__main__":
#     ops.main(SomeCharm)



# Copyright 2023,2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Project-related models for Charmcraft."""

import abc
import datetime
import pathlib
import re
import textwrap
import warnings
from collections.abc import Iterable
from typing import (
    Annotated,
    Any,
    Literal,
    cast,
)

import craft_platforms
import pydantic
import pydantic.v1
from craft_application import errors, models
from craft_application.models import PlatformsDict, VersionStr
from craft_application.util import humanize_list, safe_yaml_load
from craft_cli import CraftError, emit
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Self, override

from charmcraft import const, preprocess
from charmcraft.const import (
    SUPPORTED_BASE_STRINGS,
    BaseStr,
    BuildBaseStr,
)
from charmcraft.models import charmcraft
from charmcraft.models.charmcraft import (
    AnalysisConfig,
    BasesConfiguration,
    Charmhub,
    Links,
)
from charmcraft.parts import process_part_config

CharmcraftSummaryStr = Annotated[
    str,
    models.SummaryStr,
    pydantic.StringConstraints(max_length=200),
    # Maximum length was set to 200 characters because the 78 character maximum
    # inherited from craft-application is too restrictive, as several hundred charms
    # already exceed this maximum.
    # Eventually this limit will be reduced, ideally to 78 characters, though that may
    # never happen entirely. Reductions will only occur on major releases.
    # https://github.com/canonical/charmcraft/issues/1598
]


def _validate_field_name(name: str) -> str:
    """Check that a Juju field name matches Juju's requirements for said field name.

    See the code at:
    https://github.com/juju/juju/blob/60c6895e8ffb9b9952d7971cbdd6df810a572410/mongo/utils/validfield.go#L13-L26
    """
    if not name:
        raise ValueError("A field name cannot be empty.")
    if name.startswith("$"):
        raise ValueError("A field name cannot start with '$'.")
    if "." in name:
        raise ValueError("A field name cannot contain '.'.")
    return name


FieldName = Annotated[str, _validate_field_name]


def get_charm_file_platform_str(bases: Iterable[charmcraft.Base]) -> str:
    """Get the "platform" section of a charm file name from an iterable of bases."""
    base_strings = []
    for base in bases:
        name = base.name
        version = base.channel
        architectures = "-".join(base.architectures)
        base_strings.append(f"{name}-{version}-{architectures}")
    return "_".join(base_strings)


class CharmLib(models.CraftBaseModel):
    """A Charm library dependency for this charm."""

    lib: str = pydantic.Field(
        title="Library Path (e.g. my-charm.my_library)",
        pattern=r"[a-z][a-z0-9_-]+\.[a-z][a-z0-9_]+",
    )
    version: str = pydantic.Field(
        title="Version filter for the charm. Either an API version or a specific [api].[patch].",
        pattern=r"[0-9]+(\.[0-9]+)?",
        coerce_numbers_to_str=False,
        strict=True,
    )

    @pydantic.field_validator("lib", mode="before")
    def _validate_name(cls, value: str) -> str:
        """Validate the lib field, providing a useful error message on failure."""
        charm_name, _, lib_name = str(value).partition(".")
        if not charm_name or not lib_name:
            raise ValueError(
                f"Library name invalid. Expected '[charm_name].[lib_name]', got {value!r}"
            )
        # Accept python-importable charm names, but convert them to store-accepted names.
        if "_" in charm_name:
            charm_name = charm_name.replace("_", "-")
        if not re.fullmatch("[a-z0-9_-]+", charm_name):
            raise ValueError(
                f"Invalid charm name for lib {value!r}. Value {charm_name!r} is invalid."
            )
        if not re.fullmatch("[a-z0-9_]+", lib_name):
            raise ValueError(
                f"Library name {lib_name!r} is invalid. Library names must be valid Python module names."
            )
        return f"{charm_name}.{lib_name}"

    @pydantic.field_validator("version", mode="before")
    def _validate_api_version(cls, value: str) -> str:
        """Validate the API version field, providing a useful error message on failure."""
        api, *_ = str(value).partition(".")
        try:
            int(api)
        except ValueError:
            raise ValueError(
                f"API version not valid. Expected an integer, got {api!r}"
            ) from None
        return str(value)

    @pydantic.field_validator("version", mode="before")
    def _validate_patch_version(cls, value: str | float) -> str:
        """Validate the optional patch version, providing a useful error message."""
        if not isinstance(value, str):
            raise ValueError("Input should be a valid string")  # noqa: TRY004
        api, separator, patch = value.partition(".")
        if not separator:
            return value
        try:
            int(patch)
        except ValueError:
            raise ValueError(
                f"Patch version not valid. Expected an integer, got {patch!r}"
            ) from None
        return value

    @property
    def api_version(self) -> int:
        """The API version needed for this library."""
        return int(self.version.partition(".")[0])

    @property
    def patch_version(self) -> int | None:
        """The patch version needed for this library, or None if no patch version is specified."""
        api, _, patch = self.version.partition(".")
        if not patch:
            return None
        return int(patch)


class CharmcraftProject(models.Project, metaclass=abc.ABCMeta):
    """A craft-application compatible version of a Charmcraft project.

    This is a Project definition for charmcraft commands that are run through
    craft-application rather than the legacy charmcraft entrypoint. Eventually
    it will be the only form of the project.

    This inherits from CraftBaseModel rather than from the base craft-application Project
    in order to preserve field order. It's registered as a virtual child class below.
    """

    type: Literal["charm"]
    title: models.ProjectTitle | None = None
    summary: CharmcraftSummaryStr | None = None
    description: str | None = None

    analysis: AnalysisConfig | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            How analysis done on the charm will behave.

            Currently the only options are to ignore attributes or linters."""
        ),
    )
    charmhub: Charmhub | None = pydantic.Field(
        default=None,
        description="(DEPRECATED): Configuration for accessing charmhub.",
        deprecated=(
            "The 'charmhub' field is deprecated and no longer used. It will be removed in a "
            f"future release. Use the ${const.STORE_API_ENV_VAR}, ${const.STORE_STORAGE_ENV_VAR} "
            f"and ${const.STORE_REGISTRY_ENV_VAR} environment variables instead."
        ),
    )

    # Default project properties that Charmcraft currently does not use. Types are set
    # to be Optional[None], preventing them from being used, but allow them to be used
    # by the application.

    # Allow setting this - we don't do anything with it though, so we don't show it in the schema.
    version: SkipJsonSchema[VersionStr | None] = None
    license: SkipJsonSchema[None] = None  # pyright: ignore[reportIncompatibleVariableOverride]
    # These are inside the "links" child model.
    contact: SkipJsonSchema[None] = None  # pyright: ignore[reportIncompatibleVariableOverride]
    issues: SkipJsonSchema[None] = None  # pyright: ignore[reportIncompatibleVariableOverride]
    source_code: SkipJsonSchema[None] = None  # pyright: ignore[reportIncompatibleVariableOverride]

    charm_libs: list[CharmLib] = pydantic.Field(
        default_factory=list, title="List of libraries to use for this charm"
    )

    # These private attributes are not part of the project model but are attached here
    # because Charmcraft uses this metadata.
    _started_at: datetime.datetime = pydantic.PrivateAttr(
        default_factory=lambda: datetime.datetime.now(tz=datetime.timezone.utc)
    )
    _valid: bool = pydantic.PrivateAttr(default=False)

    @property
    def started_at(self) -> datetime.datetime:
        """Get the time that Charmcraft started running."""
        return self._started_at

    @classmethod
    def unmarshal(cls, data: dict[str, Any]):
        """Create a Charmcraft project from a dictionary of data."""
        if cls is not CharmcraftProject:
            return cls.model_validate(data)
        project_type = data.get("type")
        if project_type == "charm":
            if "bases" in data:
                return BasesCharm.unmarshal(data)
            return PlatformCharm.unmarshal(data)
        if project_type == "bundle":
            raise errors.CraftValidationError(
                "Invalid 'charmcraft.yaml' file",
                details="Support for 'type: bundle' was removed in Charmcraft 4",
                resolution="To create charm bundles, use Charmcraft 3",
                reportable=False,
                docs_url="https://documentation.ubuntu.com/charmcraft/stable/reference/files/charmcraft-yaml-file",
            )
        raise ValueError(f"field type cannot be {project_type!r}")

    @classmethod
    @override
    def from_yaml_data(cls, data: dict[str, Any], filepath: pathlib.Path) -> Self:
        return cls.unmarshal(data)

    @classmethod
    def from_yaml_file(cls, path: pathlib.Path) -> Self:
        """Instantiate this model from a YAML file.

        For use with craft-application.
        """
        try:
            with path.open() as file:
                data = safe_yaml_load(file)
        except FileNotFoundError:
            raise CraftError(f"Could not find charmcraft.yaml at '{path}'")
        except OSError as exc:
            raise CraftError(
                f"Error parsing charmcraft.yaml at '{path}'", details=exc.strerror
            ) from exc

        if not isinstance(data, dict):
            raise errors.CraftValidationError(
                "Invalid 'charmcraft.yaml' file",
                details=f"File generated a {type(data)} object, expected a dictionary",
                resolution="Ensure 'charmcraft.yaml' is valid",
                reportable=False,
                docs_url="https://juju.is/docs/sdk/charmcraft-yaml",
            )

        project_dir = path.parent

        preprocess.add_default_parts(data)
        preprocess.add_metadata(project_dir, data)
        preprocess.add_config(project_dir, data)
        preprocess.add_actions(project_dir, data)

        try:
            project = cls.unmarshal(data)
        except pydantic.ValidationError as err:
            raise errors.CraftValidationError.from_pydantic(err, file_name=path.name)
        except ValueError as err:
            error_str = "\n".join(f"- {arg}" for arg in err.args)
            raise errors.CraftValidationError(
                f"Bad charmcraft.yaml content:\n{error_str}",
            )

        return project

    @pydantic.model_validator(mode="before")
    @classmethod
    def _preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Preprocess any values that charmcraft infers, before attribute validation."""
        if "type" not in values:
            raise ValueError("Project type must be declared in charmcraft.yaml.")

        return values

    @pydantic.field_validator("platforms", mode="before")
    @classmethod
    def _preprocess_platforms(
        cls, values: dict[str, dict[str, list[str]] | None]
    ) -> dict[str, dict[str, list[str]]]:
        """Expand the dictionary into the real platforms."""
        platforms = {
            name: value
            if value
            else {
                "build-on": [name],
                "build-for": [name],
            }
            for name, value in values.items()
        }
        for name, value in platforms.items():
            if value.get("build-for") is None:
                value["build-for"] = [name]
        return platforms

    @pydantic.field_validator("parts", mode="before")
    @classmethod
    def _preprocess_parts(
        cls, parts: dict[str, dict[str, Any]] | None, info: pydantic.ValidationInfo
    ) -> dict[str, dict[str, Any]]:
        """Preprocess parts object for a charm, creating an implicit part if needed."""
        if parts is not None and not isinstance(parts, dict):
            raise TypeError(
                "'parts' in charmcraft.yaml must conform to the charmcraft.yaml spec."
            )
        if not parts:
            if "type" in info.data:
                parts = {info.data["type"]: {"plugin": info.data["type"]}}
            else:
                parts = {}
        for name, part in parts.items():
            if not isinstance(part, dict):
                raise TypeError(f"part {name!r} must be a dictionary.")
            # implicit plugin fixup
            if "plugin" not in part:
                part["plugin"] = name

        for name, part in parts.items():
            if name == "charm" and part["plugin"] == "charm":
                part.setdefault("source", ".")
        return {name: process_part_config(part) for name, part in parts.items()}

    @pydantic.model_validator(mode="after")
    def _warn_charmhub_deprecated(self) -> Self:
        repeat = False
        with warnings.catch_warnings(record=True) as caught:
            if self.charmhub:
                repeat = True
                for warning in caught:
                    if isinstance(warning.message, Warning):
                        message = warning.message.args[0]
                    else:
                        message = warning.message
                    emit.progress(f"WARNING: {message}", permanent=True)
        if repeat:
            for warning in caught:
                warnings.warn(warning.message, stacklevel=1)
        return self


class CharmProject(CharmcraftProject):
    """A base class for all charm types."""

    type: Literal["charm"]
    """The type of project. Must be the string ``charm``."""
    name: models.ProjectName = pydantic.Field(
        description=textwrap.dedent(
            """\
            The name of the project on Charmhub.

            This value will be used both in the URL of the charm on Charmhub and when
            deploying the charm with juju.

            Charms should follow the
            `charm naming guidelines <https://juju.is/docs/sdk/naming>`_."""
        ),
        examples=[
            "mysql",
            "mysql-k8s",
        ],
    )
    summary: CharmcraftSummaryStr = pydantic.Field(  # pyright: ignore[reportGeneralTypeIssues]
        description="A brief (one-line) summary of your charm.",
    )
    description: str = pydantic.Field(  # pyright: ignore[reportGeneralTypeIssues]
        description="A multi-line summary of your charm."
    )

    actions: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Defines one or more actions.

            This key is equivalent to the
            `actions.yaml file <https://juju.is/docs/sdk/actions-yaml>`_."""
        ),
    )
    assumes: list[str | dict[str, list | dict]] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Explicitly state features a Juju model must be able to provide for a
            successful deployment of this charm. When a charm includes such
            requirements, Juju performs a pre-deployment check and displays
            user-friendly error messages if a feature requirement cannot be met by the
            model that the user is trying to deploy the charm to. If the assumes
            section of the charm metadata is omitted, Juju will make a best-effort
            attempt to deploy the charm, and users must rely on the output of
            ``juju status`` to figure out whether the deployment was successful.

            The key consists of a list of features that can be given either directly
            or, depending on the complexity of the condition you want to enforce,
            nested under one or both of the boolean expressions any-of or all-of,
            as shown below. In order for a charm to be deployed, all entries in the
            assumes block must be satisfied.

            Structure::

                assumes:
                  - <feature-1>
                  - any-of:
                    - <feature-2>
                    - <feature-3>
                  - all-of:
                    - <feature-4>
                    - <feature-5>

            Juju version requirements can be specified with a string such as
            ``juju >= 3.5`` or ``juju < 4.0``. A full list of supported features
            can be found in the
            `Juju documentation <https://juju.is/docs/juju/supported-features>`_.
            """
        ),
    )
    charm_user: Literal["root", "sudoer", "non-root"] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Specifies that the charm code does not need to be run as root. Possible
            values are:

            - ``root``: the charm will run as root
            - ``sudoer``: the charm will run as a non-root user with access to root
              privileges using ``sudo``.
            - ``non-root``: the charm will run as a non-root user without ``sudo``.

            Only affects Kubernetes charms on Juju 3.6.0 or later. If not specified,
            Juju will use
            `its default behaviour <https://juju.is/docs/sdk/metadata-yaml#heading--charm-user>`_.
            """
        ),
    )
    containers: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Define a map of containers to be created adjacent to the charm (as a
            sidecar, in the same pod).

            This is required for Kubernetes charms.

            This key consists of a dictionary mapping container names to their
            specifications. Each container can be specified in terms of ``resource``,
            ``bases`` and ``mounts``, where one of either the ``resource`` or the
            ``bases`` subkeys must be defined and ``mounts`` is optional.

            - ``resource`` is the name of an OCI image resource used to create the
              container (that you will then define further in the resources block).
            - ``bases`` is a list of bases to be used for resolving a container image,
              in descending order of preference. To use it, specify a base name (for
              example, ``ubuntu`` or ``almalinux``), a ``channel`` and an
              ``architecture``.
            - ``mounts`` is a list of mounted storage volumes for this container. To
              use it, specify the name of the storage to mount from the charm
              storage and, optionally, the location where to mount the storage.

            Structure::

                containers:
                  <container name>:
                    resource: <resource name>
                    bases:
                      - name: <base name>
                        channel: <track[/risk][/branch]>
                        architectures:
                          - <architecture>
                    mounts:
                      - storage: <storage name>
                        location: <path>"""
        ),
        examples=[
            {
                "super-app": {
                    "resource": "super-app-image",
                    "mounts": [{"storage": "logs", "location": "/logs"}],
                }
            }
        ],
    )
    devices: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Devices the charm needs.

            Structure::

                devices:
                  <device name>:
                    type: gpu | nvidia.com/gpu | amd.com/gpu
                    description: <Optional description>
                    countmin: <Optional minimum number requested>
                    countmax: <Optional maximum number requested>"""
        ),
        examples=[
            {
                "amd-gpu": {
                    "type": "amd.com/gpu",
                    "description": "Some sweet AMD GPU",
                    "countmin": 1,
                    "countmax": 100,
                },
                "nvidia-gpu": {
                    "type": "nvidia.com/gpu",
                    "description": "Some NVIDIA GPUs",
                    "countmin": 20,
                },
            },
            {
                "gpus": {
                    "type": "gpu",
                    "description": "A bunch of GPUs",
                    "countmin": 2,
                    "countmax": 40,
                }
            },
        ],
    )
    extra_bindings: dict[str, Any] | None = pydantic.Field(
        default=None,
        description="A key-only mapping representing extra bindings needed.",
    )
    peers: dict[FieldName, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A map of peer relations.

            Structure::

                peers:
                  <endpoint name>:
                    interface: <Required interface name>
                    limit: <Optional: maximum number of supported connections
                    optional: <Informational only - whether the relation is required.>
                    scope: <"global" or "container" - the relation scope.>

            For more information, see
            `the Juju documentation <https://juju.is/docs/juju/relation>`_."""
        ),
        examples=[
            {
                "friend": {
                    "interface": "life",
                    "limit": 150,
                    "optional": True,
                    "scope": "container",
                }
            }
        ],
    )
    provides: dict[FieldName, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A map of interfaces this charm provides.

            Structure::

                provides:
                  <endpoint name>:
                    interface: <Required interface name>
                    limit: <Optional: maximum number of supported connections
                    optional: <Informational only - whether the relation is required.>
                    scope: <"global" or "container" - the relation scope.>

            For more information, see
            `the Juju documentation <https://juju.is/docs/juju/relation>`_."""
        ),
        examples=[{"self": {"interface": "identity"}}],
    )
    requires: dict[FieldName, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A map of relations this charm requires.

            Structure::

                requires:
                  <endpoint name>:
                    interface: <Required interface name>
                    limit: <Optional: maximum number of supported connections
                    optional: <Informational only - whether the relation is required.>
                    scope: <"global" or "container" - the relation scope.>

            For more information, see
            `the Juju documentation <https://juju.is/docs/juju/relation>`_."""
        ),
        examples=[
            {
                "parent": {
                    "interface": "birth",
                    "limit": 2,
                    "optional": False,
                    "scope": "global",
                }
            }
        ],
    )
    resources: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A mapping of resources that accompany the charm.

            See first: `Juju | Charm resource <https://juju.is/docs/juju/charm-resource>`_

            Each resource is made available when the charm is deployed. NOTE:
            Kubernetes charms must declare an ``oci-image`` type resource for each
            container declared in ``containers``.

            Structure::

                # (Optional) Additional resources that accompany the charm
                resources:
                  <resource name>:
                    # (Required) The type of the resource
                    type: file | oci-image

                    # (Optional) Description of the resource and its purpose
                    description: <description>

                    # (Required: when type:file) The filename of the resource as it
                    # should appear in the filesystem.
                    filename: <filename>"""
        ),
        examples=[
            {"water": {"type": "file", "filename": "/dev/h2o"}},
            {"super-app-image": {"type": "oci-image", "description": "My app!"}},
        ],
    )
    storage: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Storage devices requested by the charm.

            Structure::

                storage:
                  # Each key represents the name of the storage
                  <storage name>:

                    # (Required) Type of the requested storage
                    type: filesystem | block

                    # (Optional) Description of the storage requested
                    description: <description>

                    # (Optional) The mount location for filesystem stores. For
                    # multi-stores the location acts as the parent directory for each
                    # mounted store.
                    location: <location>

                    # Indicates if the storage should be made read-only (where
                    # possible). Defaults to false
                    read-only: true | false

                    # (Optional) The number of storage instances to be requested
                    multiple:
                      range: <n> | <n>-<m> | <n>- | <n>+

                    # (Optional) Minimum size of requested storage in forms G, GiB, GB.
                    # Size multipliers are M, G, T, P, E, Z or Y. With no multiplier
                    # supplied, M is implied.
                    minimum-size: <n> | <n><multiplier>

                    # (Optional) List of properties, only supported value is "transient"
                    properties:
                      - transient
            """
        ),
        examples=[
            {
                "jbod": {
                    "type": "block",
                    "description": "A block storage to use as swap space",
                    "properties": ["transient"],
                },
            },
        ],
    )
    subordinate: bool | None = pydantic.Field(
        default=None,
        description="Optional boolean to declare the charm subordinate.",
        examples=[True],
    )
    terms: list[str] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A list of terms to which the user agree by using the charm.
            These terms are not enforced by the charm, Juju or Canonical."""
        ),
        examples=[
            "Post cat pictures on Mastodon",
            "Tag your cat pictures with #caturday",
        ],
    )
    links: Links | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            (Recommended) Links to various additional information used by Charmhub."""
        ),
        examples=[
            {
                "contact": "Please send your answer to Old Pink, care of the Funny Farm, Chalfont",
                "documentation": "https://discourse.charmhub.io/t/traefik-k8s-docs-index/10778",
                "issues": "https://github.com/canonical/traefik-k8s-operator/issues",
                "source": "https://github.com/canonical/traefik-k8s-operator",
                "website": "https://charmed-kubeflow.io/",
            }
        ],
    )
    config: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            One or more configuration options for your charm.

            Structure::

                config:
                  options:
                    # Each option name is the name by which the charm will query the option.
                    <option name>:
                      # (Required) The type of the option
                      type: string | int | float | boolean | secret
                      # (Optional) The default value of the option
                      default: <a reasonable default value of the same type as the option>
                      # (Optional): A string describing the option. Also appears on charmhub.io
                      description: <description string>"""
        ),
        examples=[
            {
                "options": {
                    "name": {
                        "default": "Wiki",
                        "description": "The name or title of the Wiki",
                        "type": "string",
                    },
                    "skin": {
                        "default": "vector",
                        "description": "Skin to use for the wiki",
                        "type": "string",
                    },
                },
            },
        ],
    )


def _check_base_is_legacy(base: charmcraft.BaseDict) -> bool:
    """Check that the given base is a legacy base, usable with 'bases'."""
    base_str = f"{base['name']}@{base['channel']}"  # pyright: ignore[reportTypedDictNotRequiredAccess]
    return base_str in const.LEGACY_BASES


def _validate_base(
    base: charmcraft.BaseDict | charmcraft.LongFormBasesDict,
) -> charmcraft.LongFormBasesDict:
    if "name" in base:  # Convert short form to long form
        base = cast(
            charmcraft.LongFormBasesDict, {"build-on": [base], "run-on": [base]}
        )
    else:  # Cast to long form since we know it is one.
        base = cast(charmcraft.LongFormBasesDict, base)

    # Ensure we're only allowing legacy bases.
    for build_base in base["build-on"]:
        if not _check_base_is_legacy(build_base):
            raise ValueError(f"Base requires 'platforms' definition: {build_base}")
    for run_base in base["run-on"]:
        if not _check_base_is_legacy(run_base):
            raise ValueError(f"Base requires 'platforms' definition: {run_base}")
    return base


class BasesCharm(CharmProject):
    """A charm using the deprecated ``bases`` keyword.

    This type of charm only supports the following bases:
        - Ubuntu 18.04
        - Ubuntu 20.04
        - Ubuntu 22.04
        - Alma Linux 9
    """

    # For bases charms, accept anything so we can use platforms internally, but
    # exclude it from serialization and the JSON schema.
    platforms: SkipJsonSchema[Any] = pydantic.Field(
        default=None,
        exclude=True,
        repr=False,
    )

    # This is defined this way because using conlist makes mypy sad and using
    # a ConstrainedList child class has pydantic issues. This appears to be
    # solved with Pydantic 2.
    bases: list[
        Annotated[BasesConfiguration, pydantic.BeforeValidator(_validate_base)]
    ] = pydantic.Field(min_length=1)

    base: None = None

    parts: dict[str, dict[str, Any]] = pydantic.Field(
        default={"charm": {"plugin": "charm", "source": "."}},
        description=textwrap.dedent(
            """\
            Configures the various mechanisms to obtain, process and prepare data from
            different sources that end up being a part of the final charm.

            Keys are user-defined part names. The value of each key is a map where keys
            are part names.

            Example::

                parts:
                  libs:
                    plugin: dump
                    source: /usr/local/lib/
                    organize:
                      "libxxx.so*": lib/
                    prime:
                      - lib/""",
        ),
    )


class PlatformCharm(CharmProject):
    """Model for defining a charm using Platforms."""

    # Silencing pyright because it complains about missing default value
    base: BaseStr | None = None
    build_base: BuildBaseStr | None = None
    platforms: PlatformsDict

    parts: dict[str, dict[str, Any]] = pydantic.Field(
        description=textwrap.dedent(
            """\
            Configures the various mechanisms to obtain, process and prepare data from
            different sources that end up being a part of the final charm.

            Keys are user-defined part names. The value of each key is a map where keys
            are part names.

            Example::

                parts:
                  libs:
                    plugin: dump
                    source: /usr/local/lib/
                    organize:
                      "libxxx.so*": lib/
                    prime:
                      - lib/""",
        ),
        min_length=1,
    )

    @pydantic.model_validator(mode="after")
    def _validate_dev_base_needs_build_base(self) -> Self:
        if not self.build_base and self.base in const.DEVEL_BASE_STRINGS:
            recommended_base = (
                self.base if self.base in SUPPORTED_BASE_STRINGS else "ubuntu@devel"
            )
            raise ValueError(
                f"Base {self.base} requires a build-base (recommended: 'build-base: {recommended_base}')"
            )
        return self

    @override
    @pydantic.field_validator("platforms", mode="before")
    @classmethod
    def _populate_platforms(cls, platforms: dict[str, Any]) -> dict[str, Any]:
        """Overrides the validator to prevent platforms from being modified.

        Modifying the platforms field can break multi-base builds."""
        return platforms

    @pydantic.field_validator("parts", mode="before")
    @classmethod
    def _validate_removed_questing_plugins(
        cls, value: dict[str, dict[str, Any]], info: pydantic.ValidationInfo
    ) -> dict[str, dict[str, Any]]:
        """Check that the charm and reactive plugins aren't used on Ubuntu 25.10+."""
        plugins = {v.get("plugin", k) for k, v in value.items()}
        if not plugins & {"charm", "reactive"}:
            return value
        if (base := info.data.get("base")) in const.CHARM_OR_REACTIVE_BASES:
            return value
        if base is not None:
            raise ValueError(
                f"Cannot use 'charm' or 'reactive' plugins with base {base!r}"
            )
        # Multi-base charms.
        build_bases = {
            str(info.build_base)
            for info in craft_platforms.charm.get_platforms_charm_build_plan(
                base=None,
                platforms=pydantic.TypeAdapter(PlatformsDict).dump_python(
                    info.data["platforms"], mode="json", by_alias=True
                ),
            )
        }
        if invalid_bases := build_bases - const.CHARM_OR_REACTIVE_BASES:
            if len(invalid_bases) == 1:
                raise ValueError(
                    f"Cannot use 'charm' or 'reactive' plugins with base {invalid_bases.pop()!r}"
                )
            invalid_bases_str = humanize_list(sorted(invalid_bases), conjunction="or")
            raise ValueError(
                f"Cannot use 'charm' or 'reactive' plugins with bases {invalid_bases_str}"
            )
        return value


Charm = PlatformCharm | BasesCharm



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Service class for packing."""

from __future__ import annotations

import json
import os
import pathlib
import shutil
from collections.abc import Iterable
from typing import TYPE_CHECKING, cast

import craft_platforms
import yaml
from craft_application import services
from craft_cli import emit

import charmcraft
from charmcraft import const, errors, models, utils
from charmcraft.models import lint
from charmcraft.models.manifest import Attribute, Manifest
from charmcraft.models.metadata import CharmMetadata
from charmcraft.models.project import (
    BasesCharm,
    PlatformCharm,
)

if TYPE_CHECKING:
    from charmcraft.services.analysis import AnalysisService


class PackageService(services.PackageService):
    """Business logic for creating packages."""

    def pack(self, prime_dir: pathlib.Path, dest: pathlib.Path) -> list[pathlib.Path]:
        """Create one or more packages as appropriate.

        :param prime_dir: Directory path to the prime directory.
        :param dest: Directory into which to write the package(s).
        :returns: A list of paths to created packages.
        """
        return [self.pack_charm(prime_dir, dest)]

    def pack_charm(
        self, prime_dir: pathlib.Path, dest_dir: pathlib.Path
    ) -> pathlib.Path:
        """Pack a prime directory as a charm for a given set of bases."""
        charm_name = self.get_charm_name()
        charm_path = dest_dir / charm_name
        emit.progress(f"Packing charm {charm_name}")
        utils.build_zip(charm_path, prime_dir)

        return charm_path

    def get_charm_name(self) -> str:
        """Get a charm file name for the appropriate set of run-on bases."""
        name = self._services.get("project").get().name
        platform = self._services.get("build_plan").plan()[0].platform
        platform = platform.replace(":", "-")
        return f"{name}_{platform}.charm"

    @property
    def metadata(self) -> CharmMetadata:
        """Metadata model for this project."""
        return CharmMetadata.from_charm(
            cast("BasesCharm | PlatformCharm", self._services.get("project").get())
        )

    def _write_file_or_object(
        self, model: dict | None, filename: str, dest_dir: pathlib.Path
    ) -> None:
        """Write a yaml file to the destination directory if the given object is not None.

        This function prefers copying the file over, but will generate YAML from the given
        model otherwise.

        :param model: The dictionary to write (or None)
        :param filename: The name of the file to copy or write.
        :param dest_dir: The path of the destination directory.
        """
        if not model:
            return
        dest_dir.mkdir(parents=True, exist_ok=True)
        source_path = (
            self._services.get("project").resolve_project_file_path().parent / filename
        )
        dest_path = dest_dir / filename
        if source_path.is_file():
            shutil.copyfile(source_path, dest_path)
            return
        with dest_path.open("wt+") as dest_file:
            yaml.safe_dump(model, dest_file)

    def get_manifest(self, lint_results: Iterable[lint.CheckResult]) -> Manifest:
        """Get the manifest for this charm."""
        attributes = [
            Attribute(name=result.name, result=result.result)
            for result in lint_results
            if result.check_type == lint.CheckType.ATTRIBUTE
        ]

        if image_info := os.getenv(const.IMAGE_INFO_ENV_VAR):
            try:
                image_info = json.loads(image_info)
            except json.decoder.JSONDecodeError as exc:
                msg = f"Failed to parse the content of {const.IMAGE_INFO_ENV_VAR} environment variable"
                raise errors.CraftError(msg) from exc

        bases = self.get_manifest_bases()

        return Manifest(
            charmcraft_version=charmcraft.__version__,
            charmcraft_started_at=str(
                self._services.get("state").get("charmcraft", "started_at")
            ),
            analysis={"attributes": attributes},
            image_info=image_info,
            bases=bases,
        )

    def get_manifest_bases(self) -> list[models.Base]:
        """Get the bases used for a charm manifest from the project."""
        project = cast(
            "BasesCharm | PlatformCharm", self._services.get("project").get()
        )
        build_item = self._services.get("build_plan").plan()[0]
        if isinstance(project, BasesCharm):
            run_on_bases = []
            for project_base in project.bases:
                for build_base in project_base.build_on:
                    if build_base.name != build_item.build_base.distribution:
                        continue
                    if build_base.channel != build_item.build_base.series:
                        continue
                    if build_item.build_on not in build_base.architectures:
                        continue
                    run_on_bases.extend(project_base.run_on)
            if not run_on_bases:
                raise RuntimeError("Could not determine run-on bases.")
            return run_on_bases
        if isinstance(project, PlatformCharm):
            archs = [str(build_item.build_for)]

            # single base recipes will have a base
            if project.base:
                return [models.Base.from_str_and_arch(project.base, archs)]

            # multi-base recipes may have the base in the platform name
            platform_label = build_item.platform
            if base := craft_platforms.parse_base_and_name(platform_label)[0]:
                return [
                    models.Base(
                        name=base.distribution,
                        channel=base.series,
                        architectures=archs,
                    )
                ]

            # Otherwise, retrieve the build-for base from the platform in the project.
            # This complexity arises from building on devel bases - the BuildInfo
            # contains the devel base and not the compatibility base.
            platform = project.platforms.get(platform_label)
            if platform and platform.build_for:
                if base := craft_platforms.parse_base_and_architecture(
                    platform.build_for[0]
                )[0]:
                    return [
                        models.Base(
                            name=base.distribution,
                            channel=base.series,
                            architectures=archs,
                        )
                    ]

        raise TypeError(f"Unknown charm type {project.__class__}, cannot get bases.")

    def write_metadata(self, path: pathlib.Path) -> None:
        """Write additional charm metadata.

        :param path: The path to the prime directory.
        """
        project = cast(
            "BasesCharm | PlatformCharm", self._services.get("project").get()
        )
        path.mkdir(parents=True, exist_ok=True)
        if isinstance(project, BasesCharm | PlatformCharm):
            if project.analysis:
                ignore_checkers = {
                    *project.analysis.ignore.linters,
                    *project.analysis.ignore.attributes,
                }
            else:
                ignore_checkers = set()
            svc = cast("AnalysisService", self._services.get("analysis"))
            lint_results = svc.lint_directory(
                self._services.get("lifecycle").prime_dir, ignore=ignore_checkers
            )
            manifest = self.get_manifest(lint_results)
            # Converting the manifest to a dictionary here is fairly fragile.
            # We need to include unset/default values in order to ensure that the
            # architecture is included on each base in the manifest, even when the
            # architectures are inferred. However, we also need to exclude Nones so that
            # image-info isn't included in manifest.yaml if it doesn't exist.
            # Tread carefully when changing this next line. Treat it like an antique
            # crystal wine glass.
            (path / "manifest.yaml").write_text(
                utils.dump_yaml(
                    manifest.model_dump(
                        mode="json",
                        by_alias=True,
                        exclude_unset=False,
                        exclude_none=True,
                    )
                )
            )

        project_dict = project.marshal()

        # If there is a reactive part, defer to it for the existence of metadata.yaml.
        plugins = {
            part.get("plugin") or name  # NOTE: Not the same as part.get("plugin", name)
            for name, part in project.parts.items()
        }
        is_reactive = "reactive" in plugins
        stage_dir = self._services.get("lifecycle").project_info.dirs.stage_dir
        if is_reactive and (stage_dir / const.METADATA_FILENAME).exists():
            emit.debug(
                f"{const.METADATA_FILENAME!r} generated by charm. Not using original project metadata."
            )
        else:
            self._write_file_or_object(
                self.metadata.marshal(), const.METADATA_FILENAME, path
            )
        if is_reactive and (stage_dir / const.JUJU_ACTIONS_FILENAME).exists():
            emit.debug(
                f"{const.JUJU_ACTIONS_FILENAME!r} generated by charm. Skipping generation."
            )
        elif actions := cast(dict | None, project_dict.get("actions")):
            self._write_file_or_object(actions, "actions.yaml", path)
        if is_reactive and (stage_dir / const.JUJU_CONFIG_FILENAME).exists():
            emit.debug(
                f"{const.JUJU_CONFIG_FILENAME!r} generated by charm. Skipping generation."
            )
        elif config := cast(dict | None, project_dict.get("config")):
            self._write_file_or_object(config, "config.yaml", path)



#!/usr/bin/env python3
#
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Creation of schema for charmcraft.yaml."""

import json

from charmcraft.models.project import PlatformCharm, BasesCharm
from typing import Annotated, Any
import pydantic

def _charm_type_discriminator(charm_model: Any) -> str:  # noqa: ANN401
    if hasattr(charm_model, "platforms") or "platforms" in charm_model:
        return "platformcharm"
    return "basescharm"

Charm = Annotated[
    Annotated[PlatformCharm, pydantic.Tag("platformcharm")]
    | Annotated[BasesCharm, pydantic.Tag("basescharm")],
    pydantic.Discriminator(_charm_type_discriminator)
]

if __name__ == "__main__":
    print(json.dumps(pydantic.TypeAdapter(Charm).json_schema(), indent="	"))



# Copyright 2024-2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""General fixtures for integration tests."""

import pathlib
from unittest import mock

import craft_application
import craft_store
import pytest

from charmcraft import application, services
from charmcraft.application import commands


@pytest.fixture
def project_path(tmp_path: pathlib.Path):
    path = tmp_path / "project"
    path.mkdir()
    return path


@pytest.fixture
def service_factory(
    new_path: pathlib.Path,
    fake_project_file,
    project_path,
    monkeypatch: pytest.MonkeyPatch,
):
    services.register_services()
    factory = craft_application.ServiceFactory(app=application.APP_METADATA)
    factory.get("store").client = mock.Mock(spec_set=craft_store.StoreClient)  # pyright: ignore[reportAttributeAccessIssue]
    factory.update_kwargs(
        "charm_libs",
        project_dir=project_path,
    )
    factory.update_kwargs(
        "lifecycle",
        work_dir=new_path,
        cache_dir="~/.cache",
    )
    factory.update_kwargs(
        "project",
        project_dir=project_path,
    )
    factory.update_kwargs(
        "provider",
        work_dir=new_path,
    )
    factory.get("project").configure(
        platform=None,
        build_for=None,
    )
    monkeypatch.setenv("CRAFT_STATE_DIR", str(new_path / "state"))
    factory.get("state").set(
        "charmcraft", "started_at", value="2020-03-14T00:00:00+00:00"
    )
    return factory


@pytest.fixture
def app(monkeypatch, new_path, service_factory):
    monkeypatch.setenv("CRAFT_DEBUG", "1")
    app = application.Charmcraft(app=application.APP_METADATA, services=service_factory)
    app._configure_services(None)
    commands.fill_command_groups(app)

    return app



# Flask application that keeps track of visitors using a database

import datetime
import os

import flask
import psycopg2

app = flask.Flask(__name__)
app.config.from_prefixed_env()

DATABASE_URI = os.environ["POSTGRESQL_DB_CONNECT_STRING"]


@app.route("/")
def index():
    with psycopg2.connect(DATABASE_URI) as conn, conn.cursor() as cur:
        user_agent = flask.request.headers.get('User-Agent')
        timestamp = datetime.datetime.now()

        cur.execute(
            "INSERT INTO visitors (timestamp, user_agent) VALUES (%s, %s)",
            (timestamp, user_agent)
        )
        conn.commit()


    greeting = app.config.get("GREETING", "Hello, world!")
    return f"{greeting}\n"


@app.route("/visitors")
def visitors():
    with psycopg2.connect(DATABASE_URI) as conn, conn.cursor() as cur:
        cur.execute("SELECT COUNT(*) FROM visitors")
        total_visitors = cur.fetchone()[0]

    return f"{total_visitors}\n"


if __name__ == "__main__":
    app.run()



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for the lifecycle service."""

import craft_platforms
import distro
import pytest
from craft_application import ServiceFactory, errors, util


def test_init_lifecycle(service_factory: ServiceFactory):
    """Test the setup of a parts lifecycle, implicitly testing setup."""

    service_factory.get("lifecycle")._init_lifecycle_manager()


def test_lifecycle_build_for_invalid(
    monkeypatch: pytest.MonkeyPatch, service_factory: ServiceFactory
):
    monkeypatch.setattr(
        service_factory.get("build_plan"),
        "plan",
        lambda: [
            craft_platforms.BuildInfo(
                platform="something",
                build_on=craft_platforms.DebianArchitecture.from_host(),
                build_for="invalid",  # pyright: ignore[reportArgumentType]
                build_base=craft_platforms.DistroBase(distro.id(), distro.version()),
            )
        ],
    )

    with pytest.raises(
        errors.PartsLifecycleError, match="[Aa]rchitecture '[a-z]+' is not supported"
    ):
        service_factory.get("lifecycle")


def test_lifecycle_build_for_all(
    monkeypatch: pytest.MonkeyPatch, service_factory: ServiceFactory
):
    monkeypatch.setattr(
        service_factory.get("build_plan"),
        "plan",
        lambda: [
            craft_platforms.BuildInfo(
                platform="something",
                build_on=craft_platforms.DebianArchitecture.from_host(),
                build_for="all",
                build_base=craft_platforms.DistroBase(distro.id(), distro.version()),
            )
        ],
    )
    lifecycle = service_factory.get("lifecycle")

    lcm = lifecycle._init_lifecycle_manager()

    assert lcm._target_arch == util.get_host_architecture()


def test_lifecycle_build_for_multi(
    monkeypatch: pytest.MonkeyPatch, service_factory: ServiceFactory
):
    host_arch = craft_platforms.DebianArchitecture.from_host()
    arches = {"arm64", "riscv64", "amd64"} - {host_arch}
    foreign_arch = next(iter(arches))

    monkeypatch.setattr(
        service_factory.get("build_plan"),
        "plan",
        lambda: [
            craft_platforms.BuildInfo(
                platform="something",
                build_on=host_arch,
                build_for=f"{foreign_arch}-{host_arch}",  # pyright: ignore[reportArgumentType]
                build_base=craft_platforms.DistroBase(distro.id(), distro.version()),
            )
        ],
    )

    lifecycle = service_factory.get("lifecycle")

    lcm = lifecycle._init_lifecycle_manager()

    assert lcm._target_arch == foreign_arch



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for the Image service."""

import itertools
import json
from unittest import mock

import docker
import docker.errors
import docker.models.images
import pytest

from charmcraft import application, const, utils
from charmcraft.errors import SkopeoError
from charmcraft.services.image import ImageService, OCIMetadata


@pytest.fixture
def mock_docker() -> mock.Mock:
    return mock.Mock(spec_set=docker.DockerClient)


@pytest.fixture
def mock_skopeo(fake_process) -> mock.Mock:
    fake_process.register(["/skopeo", "--version"])
    return mock.Mock(wraps=utils.Skopeo(skopeo_path="/skopeo"))


@pytest.fixture
def image_service(service_factory, mock_skopeo, mock_docker) -> ImageService:
    service = ImageService(app=application.APP_METADATA, services=service_factory)
    service._skopeo = mock_skopeo
    service._docker = mock_docker
    return service


@pytest.mark.parametrize(
    ("url", "name"),
    [
        (
            "docker://hello-world@sha256:18a657d0cc1c7d0678a3fbea8b7eb4918bba25968d3e1b0adebfa71caddbc346",
            "hello-world@sha256:18a657d0cc1c7d0678a3fbea8b7eb4918bba25968d3e1b0adebfa71caddbc346",
        ),
        (
            "hello-world@sha256:18a657d0cc1c7d0678a3fbea8b7eb4918bba25968d3e1b0adebfa71caddbc346",
            "hello-world@sha256:18a657d0cc1c7d0678a3fbea8b7eb4918bba25968d3e1b0adebfa71caddbc346",
        ),
        (
            "docker://ghcr.io/canonical/charmed-mysql@sha256:89b8305613f6ce94f78a7c9b4baedef78f2816fd6bc74c00f6607bc5e57bd8e6",
            "ghcr.io/canonical/charmed-mysql@sha256:89b8305613f6ce94f78a7c9b4baedef78f2816fd6bc74c00f6607bc5e57bd8e6",
        ),
        (
            "docker://quay.io/prometheus/blackbox-exporter:v0.24.0",
            "quay.io/prometheus/blackbox-exporter:v0.24.0",
        ),
        (
            "docker://quay.io/prometheus/blackbox-exporter:v0.24.0@sha256:3af31f8bd1ad2907b4b0f7c485fde3de0a8ee0b498d42fc971f0698885c03acb",
            "quay.io/prometheus/blackbox-exporter:v0.24.0@sha256:3af31f8bd1ad2907b4b0f7c485fde3de0a8ee0b498d42fc971f0698885c03acb",
        ),
    ],
)
def test_get_name_from_url(url: str, name: str):
    assert ImageService.get_name_from_url(url) == name


@pytest.mark.parametrize(
    ("go_arch", "charm_arch"),
    [
        *(
            (key, const.CharmArch(value))
            for key, value in const.GO_ARCH_TO_CHARM_ARCH.items()
        ),
        ("amd64", "amd64"),
        ("arm64", "arm64"),
        ("riscv64", "riscv64"),
        ("s390x", "s390x"),
    ],
)
def test_convert_go_arch_to_charm_arch(go_arch: str, charm_arch: const.CharmArch):
    assert ImageService.convert_go_arch_to_charm_arch(go_arch) == charm_arch


def test_get_maybe_id_from_docker_success(image_service: ImageService, mock_docker):
    expected = "sha256:some-sha-hash"
    mock_docker.images.get.return_value = docker.models.images.Image(
        attrs={"Id": expected}
    )

    result = image_service.get_maybe_id_from_docker("some-image")

    mock_docker.images.get.assert_called_once_with("some-image")
    assert result == expected


def test_get_maybe_id_from_docker_failure(image_service: ImageService, mock_docker):
    mock_docker.images.get.side_effect = docker.errors.ImageNotFound("womp womp")

    assert image_service.get_maybe_id_from_docker("some-image") is None


def test_get_maybe_id_from_docker_no_docker(image_service: ImageService):
    image_service._docker = None

    assert image_service.get_maybe_id_from_docker("some-image") is None


@pytest.mark.parametrize("image", ["my-image"])
@pytest.mark.parametrize("architecture", const.CharmArch)
def test_inspect_single_arch(
    fake_process,
    image_service: ImageService,
    mock_skopeo,
    image: str,
    architecture,
):
    fake_process.register(
        ["/skopeo", "inspect", "--raw", image],
        stdout=json.dumps({"raw_manifest": True}),
    )
    fake_process.register(
        ["/skopeo", "inspect", image],
        stdout=json.dumps({"Digest": "Reader's", "Architecture": architecture}),
    )

    actual = image_service.inspect(image)

    assert actual == OCIMetadata(
        path=image, digest="Reader's", architectures=[architecture]
    )


@pytest.mark.parametrize("image", ["my-image"])
@pytest.mark.parametrize("architectures", itertools.product(const.CharmArch, repeat=2))
def test_inspect_two_arch(
    fake_process,
    image_service: ImageService,
    mock_skopeo,
    image: str,
    architectures,
):
    fake_process.register(
        ["/skopeo", "inspect", "--raw", image],
        stdout=json.dumps(
            {
                "manifests": [
                    {"platform": {"os": "linux", "architecture": str(arch)}}
                    for arch in architectures
                ],
            }
        ),
    )
    fake_process.register(
        ["/skopeo", "inspect", image],
        stdout=json.dumps({"Digest": "Reader's", "Architecture": "amd64"}),
    )

    actual = image_service.inspect(image)

    assert actual == OCIMetadata(
        path=image, digest="Reader's", architectures=list(architectures)
    )


@pytest.mark.parametrize("image", ["my-image"])
@pytest.mark.parametrize(
    ("stderr", "error_msg"),
    [
        (
            'time="2025-06-25T19:14:57-04:00" level=fatal msg="Error parsing image name \\"docker://ghcr.io/canonical/spark-integration-hub:3.4-22.04_edge@sha256:0b9a40435440256b1c10020bd59d19e186ea68d8973fc8f2310010f9bd4e3459\\": Docker references with both a tag and digest are currently not supported"',
            "Docker references with both a tag and digest are currently not supported",
        ),
        ("level=fatal error='No message'", "Unknown error from skopeo."),
        ("Something unparsable", "Unknown error from skopeo."),
    ],
)
def test_inspect_skopeo_error(
    fake_process,
    image_service: ImageService,
    mock_skopeo,
    image: str,
    stderr: str,
    error_msg: str,
):
    fake_process.register(
        ["/skopeo", "inspect", "--raw", image],
        stdout="",
        stderr=stderr,
        returncode=1,
    )

    with pytest.raises(SkopeoError) as exc_info:
        image_service.inspect(image)

    assert error_msg in str(exc_info.value)
